


import jk_logging

from jk_json.tools import *

from ._SchemaAST import _SchemaAST, ALWAYS_FALSE, ALWAYS_TRUE
from .schema_validator import Validator, compileAst



def _walk(v):
	if isinstance(v, (tuple, list)):
		for x in v:
			yield x
	else:
		yield v
#

def _getN(dictionary, key, expectedTypeOrTypes):
	if key in dictionary:
		v = dictionary[key]
		if isinstance(expectedTypeOrTypes, (tuple, list)):
			if type(v) in expectedTypeOrTypes:
				return v
			else:
				raise Exception("Value of unexpected type " + str(type(v)) + " at key " + key)
		else:
			if type(v) == expectedTypeOrTypes:
				return v
			else:
				raise Exception("Value of unexpected type " + str(type(v)) + " at key " + key)
	else:
		return None
#

def _getE(dictionary, key, expectedTypeOrTypes):
	if key in dictionary:
		v = dictionary[key]
		if isinstance(expectedTypeOrTypes, (tuple, list)):
			if type(v) in expectedTypeOrTypes:
				return v
			else:
				raise Exception("Value of unexpected type " + str(type(v)) + " at key " + key)
		else:
			if type(v) == expectedTypeOrTypes:
				return v
			else:
				raise Exception("Value of unexpected type " + str(type(v)) + " at key " + key)
	else:
		raise Exception("Value expected at key " + key)
#










class SchemaContext(object):

	def __init__(self):
		self.definitions = {}
	#

#



class SchemaParser(object):

	TYPESTR_TO_TYPEID_MAP = {
		"null": TYPE_NULL,
		"number": TYPE_FLOAT,
		"integer": TYPE_INT,
		"boolean": TYPE_BOOL,
		"string": TYPE_STR,
		"array": TYPE_ARRAY,
		"object": TYPE_OBJECT,
	}

	@staticmethod
	def parse(jsonDef, log = None):
		if log is None:
			log = jk_logging.NullLogger.create()
		ctx = SchemaContext()
		log.debug("Building AST ...")
		ast = SchemaParser.__parse(ctx, jsonDef)
		log.debug("Compiling AST ...")
		return Validator(compileAst(ast, log))
	#

	#
	# Perform the parsing.
	#
	# @return		_SchemaAST		An AST object.
	#
	@staticmethod
	def __parse(ctx, jsonDef):

		if jsonDef is None:
			return None

		if jsonDef == True:
			return ALWAYS_TRUE

		if jsonDef == False:
			return ALWAYS_FALSE

		if not isinstance(jsonDef, dict):
			raise Exception()

		if "definitions" in jsonDef:
			definitions = _getE(jsonDef, "definitions", dict)
			newDefs = {}
			for defName, defSchema in definitions.items():
				p = SchemaParser.__parse(ctx, defSchema)
				if p:
					newDefs[defName] = p
			if newDefs:
				ctx.definitions.update(newDefs)

		any_allowedTypes = None
		if "type" in jsonDef:
			any_allowedTypes = []
			for t in _walk(jsonDef["type"]):
				if t in SchemaParser.TYPESTR_TO_TYPEID_MAP:
					typeID = SchemaParser.TYPESTR_TO_TYPEID_MAP[t]
					any_allowedTypes.append(typeID)
					if typeID == TYPE_FLOAT:
						any_allowedTypes.append(TYPE_INT)
				else:
					raise Exception("Invalid type specified: " + repr(t))
			if len(any_allowedTypes) == 0:
				any_allowedTypes = None

		any_allowedValues = _getN(jsonDef, "enum", (tuple, list))

		any_mustBeExactValue = None
		if "const" in jsonDef:
			any_mustBeExactValue = ( jsonDef["const"], )

		number_multipleValueOf = _getN(jsonDef, "multipleOf", (int, float))

		number_maximum = _getN(jsonDef, "maximum", (int, float))

		number_exclusiveMaximum = _getN(jsonDef, "exclusiveMaximum", (int, float))

		number_minimum = _getN(jsonDef, "minimum", (int, float))

		number_exclusiveMinimum = _getN(jsonDef, "exclusiveMinimum", (int, float))

		string_minLength = _getN(jsonDef, "minLength", int)

		string_maxLength = _getN(jsonDef, "maxLength", int)

		string_pattern = None
		if "pattern" in jsonDef:
			string_pattern = _getE(jsonDef, "pattern", str)

		array_itemsMustNotBeNone = False
		array_itemsMustBeNone = False
		array_items = None
		if "items" in jsonDef:
			x = jsonDef["items"]
			if isinstance(x, dict):
				s = SchemaParser.__parse(ctx, x)
				if s.isNotEmpty():
					array_items = s
			elif isinstance(x, list):
				array_items = []
				for x in _walk(jsonDef["items"]):
					s = SchemaParser.__parse(ctx, x)
					array_items.append(s)
				#if len(array_items) == 0:
				#	array_items = None
			elif isinstance(x, bool):
				if x:
					array_itemsMustNotBeNone = True
				else:
					array_itemsMustBeNone = True
			else:
				raise Exception("Invalid value specified for 'items'")

		array_additionalItems = None
		if "additionalItems" in jsonDef:
			array_additionalItems = SchemaParser.__parse(ctx, jsonDef["additionalItems"])

		array_maxItems = _getN(jsonDef, "maxItems", int)

		array_minItems = _getN(jsonDef, "minItems", int)

		array_itemsMustBeUnique = None
		if "uniqueItems" in jsonDef:
			array_itemsMustBeUnique = _getE(jsonDef, "uniqueItems", bool)

		array_mustContainExactValue = SchemaParser.__parse(ctx, _getN(jsonDef, "contains", (dict, bool)))

		object_properties = None
		if "properties" in jsonDef:
			object_properties = {}
			for propertyName in jsonDef["properties"]:
				p = SchemaParser.__parse(ctx, jsonDef["properties"][propertyName])
				if p:
					object_properties[propertyName] = p
			if len(object_properties) == 0:
				object_properties = None

		object_patternProperties = None
		if "patternProperties" in jsonDef:
			object_patternProperties = []
			for propertyPattern in jsonDef["patternProperties"]:
				p = SchemaParser.__parse(ctx, jsonDef["patternProperties"][propertyPattern])
				if p:
					object_patternProperties.append((propertyPattern, p))
			if len(object_patternProperties) == 0:
				object_patternProperties = None

		object_maxProperties = jsonDef.get("maxProperties", None)

		object_minProperties = jsonDef.get("minProperties", None)

		object_additionalProperties = None
		if "additionalProperties" in jsonDef:
			object_additionalProperties = SchemaParser.__parse(ctx, jsonDef["additionalProperties"])

		object_requiredProperties = jsonDef.get("required", None)

		object_propertyDependencyObjectMustMatchSchema = None
		object_propertyDependencyOtherPropertyMustExist = None
		if "dependencies" in jsonDef:
			object_propertyDependencyObjectMustMatchSchema = {}
			object_propertyDependencyOtherPropertyMustExist = {}
			for propertyName in jsonDef["dependencies"]:
				v = jsonDef["dependencies"][propertyName]
				if isinstance(v, (tuple, list)):
					object_propertyDependencyOtherPropertyMustExist[propertyName] = v
				elif isinstance(v, (dict, bool)):
					p = SchemaParser.__parse(ctx, v)
					if p != None:
						object_propertyDependencyObjectMustMatchSchema[propertyName] = p
				else:
					raise Exception("Not supported!")
			if len(object_propertyDependencyObjectMustMatchSchema) == 0:
				object_propertyDependencyObjectMustMatchSchema = None
			if len(object_propertyDependencyOtherPropertyMustExist) == 0:
				object_propertyDependencyOtherPropertyMustExist = None

		object_propertyNames = None
		if "propertyNames" in jsonDef:
			object_propertyNames = SchemaParser.__parse(ctx, jsonDef["propertyNames"])

		op_if = None
		op_then = None
		op_else = None
		xIf = _getN(jsonDef, "if", dict)
		if xIf is None:
			xThen = None
			xElse = None
		else:
			xThen = _getN(jsonDef, "then", dict)
			xElse = _getN(jsonDef, "else", dict)
			if (xThen != None) or (xElse != None):
				op_if = SchemaParser.__parse(ctx, xIf)
				if xThen:
					op_then = SchemaParser.__parse(ctx, xThen)
				if xElse:
					op_else = SchemaParser.__parse(ctx, xElse)

		op_not = None
		if "not" in jsonDef:
			op_not = SchemaParser.__parse(ctx, _getE(jsonDef, "not", (dict, bool)))

		op_allOf = None
		if "allOf" in jsonDef:
			subSchemas = _getE(jsonDef, "allOf", list)
			if len(subSchemas) == 0:
				raise Exception("'allOf' has empty schema list!")
			op_allOf = []
			for x in subSchemas:
				p = SchemaParser.__parse(ctx, x)
				if p:
					op_allOf.append(p)
			#if len(op_allOf) == 0:
			#	raise Exception("'allOf' results in empty schema list!")

		op_anyOf = None
		if "anyOf" in jsonDef:
			subSchemas = _getE(jsonDef, "anyOf", list)
			if len(subSchemas) == 0:
				raise Exception("'anyOf' has empty schema list!")
			op_anyOf = []
			for x in subSchemas:
				p = SchemaParser.__parse(ctx, x)
				if p:
					op_anyOf.append(p)
			#if len(op_anyOf) == 0:
			#	raise Exception("'anyOf' results in empty schema list!")

		op_oneOf = None
		if "oneOf" in jsonDef:
			subSchemas = _getE(jsonDef, "oneOf", list)
			if len(subSchemas) == 0:
				raise Exception("'oneOf' has empty schema list!")
			op_oneOf = []
			for x in subSchemas:
				p = SchemaParser.__parse(ctx, x)
				if p:
					op_oneOf.append(p)
			#if len(op_oneOf) == 0:
			#	raise Exception("'oneOf' results in empty schema list!")

		return _SchemaAST(
			any_allowedTypes,									# list<int>
			any_allowedValues,									# list<*>
			any_mustBeExactValue,								# 1tuple<*>
			number_multipleValueOf,								# int|float
			number_maximum,										# int|float
			number_exclusiveMaximum,							# int|float
			number_minimum,										# int|float
			number_exclusiveMinimum,							# int|float
			string_minLength,									# int
			string_maxLength,									# int
			string_pattern,										# regexstr
			array_items,										# _SchemaAST[]
			array_additionalItems,								# _SchemaAST
			array_maxItems,										# int
			array_minItems,										# int
			array_itemsMustBeUnique,							# bool
			array_mustContainExactValue,						# _SchemaAST
			array_itemsMustNotBeNone,							# bool
			array_itemsMustBeNone,								# bool
			object_minProperties,								# int
			object_maxProperties,								# int
			object_properties,									# dict<str,_SchemaAST>
			object_patternProperties,							# list<(regexstr,_SchemaAST)>
			object_propertyDependencyObjectMustMatchSchema,		# dict<str,_SchemaAST>
			object_propertyDependencyOtherPropertyMustExist,	# dict<str,str[]>
			object_propertyNames,								# _SchemaAST
			object_requiredProperties,							# str[]
			object_additionalProperties,						# _SchemaAST
			op_if,												# _SchemaAST
			op_then,											# _SchemaAST
			op_else,											# _SchemaAST
			op_not,												# _SchemaAST
			op_allOf,											# list<_SchemaAST>
			op_anyOf,											# list<_SchemaAST>
			op_oneOf											# list<_SchemaAST>
		)
	#

#















#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import re
import collections






from .Token import *
from .ParserErrorException import *





_TokenizerActionTuple = collections.namedtuple("_TokenizerActionTuple", [ "actionID", "data" ])

#
# A tokenization action. Use the methods provided in `TokenizerAction` to provide actions.
#
class TokenizerAction(object):

	ERROR = 1
	ADVANCE = 2
	SWITCHMODE = 3
	EMITELEMENT = 4
	BEGINBUFFER = 5
	APPENDELEMENTTOBUFFER = 6
	EMITBUFFER = 7
	DROPBUFFER = 8
	APPENDTEXTTOBUFFER = 9

	#
	# Emit a tokenization error.
	#
	# @param	str errorCode			An error code the error message should be prefixed with. This code does not necessarily
	#									need to be unique: Multiple tokenization problems might emit the same error. (Nevertheless
	#									you should use the same error code for the same error class.)
	# @param	str errorMessage		A user friendly error message.
	# @param	str errorIdentifier		An error identifier. This ID should be unique in order to allow tracking of the
	#									error back to the tokenization rule. This is ment for debugging only.
	#
	@staticmethod
	def error(errorCode, errorMessage, errorIdentifier):
		return _TokenizerActionTuple(TokenizerAction.ERROR, errorMessage)
	#

	#
	# Advance a single character in the stream = ignore the current character.
	#
	@staticmethod
	def advance():
		return _TokenizerActionTuple(TokenizerAction.ADVANCE, None)
	#

	#
	# Switch to a differnt parsing mode.
	#
	@staticmethod
	def switchMode(mode):
		return _TokenizerActionTuple(TokenizerAction.SWITCHMODE, mode)
	#

	#
	# Emit a token. The buffer will remain untouched.
	#
	@staticmethod
	def emitElement(tokenType):
		return _TokenizerActionTuple(TokenizerAction.EMITELEMENT, tokenType)
	#

	#
	# Reset the buffer. The buffer will now be empty and new data can be written to it.
	#
	@staticmethod
	def beginBuffer():
		return _TokenizerActionTuple(TokenizerAction.BEGINBUFFER, None)
	#

	#
	# Append the current parsed item to the buffer. What exactly that is and how much data this element contains depends on the current parsing action.
	#
	@staticmethod
	def appendElementToBuffer(callbackFunction = None):
		return _TokenizerActionTuple(TokenizerAction.APPENDELEMENTTOBUFFER, callbackFunction)
	#

	#
	# Emit the buffer = create a token with the specified token type containing the current buffer's content.
	#
	@staticmethod
	def emitBuffer(tokenType):
		return _TokenizerActionTuple(TokenizerAction.EMITBUFFER, tokenType)
	#

	#
	# Reset the buffer. This is the same as "beginBuffer()".
	#
	@staticmethod
	def dropBuffer():
		return _TokenizerActionTuple(TokenizerAction.DROPBUFFER, None)
	#

	#
	# Append the specified text to the buffer.
	#
	@staticmethod
	def appendTextToBuffer(text):
		return _TokenizerActionTuple(TokenizerAction.APPENDTEXTTOBUFFER, text)
	#

#





_TokenizerPatternTuple = collections.namedtuple("_TokenizerPatternTuple", [ "patternID", "data", "dataLength" ])

#
# A tokenization pattern. Use the methods provided in `TokenizerPattern` to provide patterns.
#
class TokenizerPattern(object):

	EXACTCHAR = 1
	ANYCHAR = 2
	EXACTSTRING = 3
	REGEX = 4

	#
	# Is the character encountered equal to the specified character?
	#
	@staticmethod
	def exactChar(data):
		assert isinstance(data, str)
		assert len(data) == 1
		return _TokenizerPatternTuple(TokenizerPattern.EXACTCHAR, data, 1)
	#

	#
	# Are the characters encountered exactly the specified sequence?
	#
	@staticmethod
	def exactSequence(data):
		assert isinstance(data, str)
		assert len(data) > 0
		return _TokenizerPatternTuple(TokenizerPattern.EXACTSTRING, data, len(data))
	#

	#
	# Is the character encountered any of the specified characters?
	#
	@staticmethod
	def anyOfTheseChars(data):
		assert isinstance(data, str)
		assert len(data) > 0
		return _TokenizerPatternTuple(TokenizerPattern.ANYCHAR, data, -1)
	#

	#
	# Do the characters encountered match the specified regular expression?
	#
	@staticmethod
	def regEx(data):
		return _TokenizerPatternTuple(TokenizerPattern.REGEX, re.compile(data), -1)
	#

#




#
# This class represents a table that is part of the initialization data of a tokenizer.
#
class TokenizingTable(object):

	def __init__(self, tableID):
		self.tableID = tableID
		self.rows = []
		self.onOtherActions = None
		self.onEOSActions = None
	#

	def addPatternRow(self, pattern, actions):
		assert isinstance(pattern, _TokenizerPatternTuple)
		assert isinstance(actions, (tuple, list))
		for action in actions:
			assert isinstance(action, _TokenizerActionTuple)
		self.rows.append((pattern.patternID, pattern.data, pattern.dataLength, tuple(actions)))
	#

	def setOther(self, actions):
		assert isinstance(actions, (tuple, list))
		for action in actions:
			assert isinstance(action, _TokenizerActionTuple)
		self.onOtherActions = actions
	#

	def setEOS(self, actions):
		assert isinstance(actions, (tuple, list))
		for action in actions:
			assert isinstance(action, _TokenizerActionTuple)
		self.onEOSActions = actions
	#

#




#
# The base class or tokenizers.
#
class TokenizerBase(object):

	def __init__(self, tables):
		assert isinstance(tables, (tuple, list))
		for table in tables:
			assert isinstance(table, TokenizingTable)

		self.__tables = tuple(tables)
	#

	#
	# Create a set of empty tables.
	#
	# @param	int numberOfTables		The number of tables to create.
	# @return	tuple					A tuple containing exactly the number of tables specified.
	#
	def createTables(self, numberOfTables):
		assert isinstance(numberOfTables, int)
		assert numberOfTables > 1

		ret = []
		for n in range(0, numberOfTables):
			ret.append(TokenizingTable(n))
		return tuple(ret)
	#

	#
	# Tokenizes the specified string.
	#
	# The following token types are supported:
	# * "d" : Delimiter
	# * "i" : Integer
	# * "f" : Float
	# * "s" : String
	# * "w" : Word
	# * "eol" : New line
	# * "eos" : End of stream
	#
	# @param	str textData		The text data to tokenize.
	# @param	str sourceID		A file path or an URL that defines the origin of the source.
	# @return	iterator<Token>		Returns an iterator that provides tokens.
	#
	def tokenize(self, textData, sourceID = None, bDebuggingEnabled = False):
		i = 0
		maxi = len(textData)
		lineNo = 0
		charPos = 0
		bufferLineNo = 0
		bufferCharPos = 0

		if bDebuggingEnabled:
			print("Now tokenizing " + str(maxi) + " characters ...")

		buffer = ""
		mode = 0
		spanText = None
		nextLineNo = 0
		nextCharPos = 0

		while i < maxi:
			if bDebuggingEnabled:
				print("---- mode:" + str(mode) + " -- i:" + str(i))
			if (mode < 0) or (mode >= len(self.__tables)):
				raise Exception("Requested tokenization table does not exist: " + str(mode))
			currentTable = self.__tables[mode]

			selectedActions = None
			spanLength = None
			spanText = None
			iEnd = None
			for patternID, patternData, patternDataLength, actions in currentTable.rows:
				if patternID == TokenizerPattern.EXACTCHAR:
					if textData[i] == patternData:
						spanLength = 1
						iEnd = i + 1
						spanText = textData[i]
						selectedActions = actions
						break
				elif patternID == TokenizerPattern.ANYCHAR:
					if patternData.find(textData[i]) >= 0:
						spanLength = 1
						iEnd = i + 1
						spanText = textData[i]
						selectedActions = actions
						break
				elif patternID == TokenizerPattern.EXACTSTRING:
					iEnd = i + patternDataLength
					if (iEnd <= maxi) and (patternData == textData[i:iEnd]):
						spanLength = patternDataLength
						iEnd = i + patternDataLength
						spanText = patternData
						selectedActions = actions
						break
				elif patternID == TokenizerPattern.REGEX:
					pr = patternData.match(textData, i)
					if pr:
						iEnd = pr.end()
						spanText = textData[i:iEnd]
						spanLength = iEnd - i
						selectedActions = actions
						break
				else:
					raise Exception()

			if selectedActions is None:
				spanLength = 1
				iEnd = i + 1
				spanText = textData[i]
				selectedActions = currentTable.onOtherActions

			nextLineNo = lineNo
			nextCharPos = charPos
			for c in spanText:
				if c == "\n":
					nextLineNo += 1
					nextCharPos = 0
				else:
					nextCharPos += 1

			for actionID, actionData in selectedActions:
				if actionID == TokenizerAction.ERROR:
					raise ParserErrorException(SourceCodeLocation(sourceID, lineNo, charPos, lineNo, charPos), actionData)
				elif actionID == TokenizerAction.ADVANCE:
					lineNo = nextLineNo
					charPos = nextCharPos
					i = iEnd
				elif actionID == TokenizerAction.SWITCHMODE:
					mode = actionData
				elif actionID == TokenizerAction.EMITELEMENT:
					yield Token(actionData, spanText, sourceID, lineNo, charPos, nextLineNo, nextCharPos)
				elif actionID == TokenizerAction.BEGINBUFFER:
					bufferLineNo = lineNo
					bufferCharPos = charPos
					buffer = ""
				elif actionID == TokenizerAction.APPENDELEMENTTOBUFFER:
					if actionData != None:
						buffer += actionData(spanText)
					else:
						buffer += spanText
				elif actionID == TokenizerAction.EMITBUFFER:
					yield Token(actionData, buffer, sourceID, bufferLineNo, bufferCharPos, lineNo, charPos)
					buffer = ""
				elif actionID == TokenizerAction.DROPBUFFER:
					bufferLineNo = lineNo
					bufferCharPos = charPos
					buffer = ""
				elif actionID == TokenizerAction.APPENDTEXTTOBUFFER:
					buffer += actionData
				else:
					raise Exception()

		currentTable = self.__tables[mode]
		for actionID, actionData in currentTable.onEOSActions:
			if actionID == TokenizerAction.ERROR:
				raise ParserErrorException(SourceCodeLocation(sourceID, lineNo, charPos, lineNo, charPos), actionData)
			elif actionID == TokenizerAction.ADVANCE:
				lineNo = nextLineNo
				charPos = nextCharPos
				i = iEnd
			elif actionID == TokenizerAction.SWITCHMODE:
				mode = actionData
			elif actionID == TokenizerAction.EMITELEMENT:
				yield Token(actionData, spanText, sourceID, lineNo, charPos, nextLineNo, nextCharPos)
			elif actionID == TokenizerAction.BEGINBUFFER:
				bufferLineNo = lineNo
				bufferCharPos = charPos
				buffer = ""
			elif actionID == TokenizerAction.APPENDELEMENTTOBUFFER:
				if actionData is None:
					buffer += actionData(spanText)
				else:
					buffer += spanText
			elif actionID == TokenizerAction.EMITBUFFER:
				yield Token(actionData, buffer, sourceID, bufferLineNo, bufferCharPos, lineNo, charPos)
				buffer = ""
			elif actionID == TokenizerAction.DROPBUFFER:
				bufferLineNo = lineNo
				bufferCharPos = charPos
				buffer = ""
			elif actionID == TokenizerAction.APPENDTEXTTOBUFFER:
				buffer += actionData
			else:
				raise Exception()

		if bDebuggingEnabled:
			print("Tokenizing completed.")

	#

#













import typing
import re
import random

import jk_json
import jk_utils

from .re import compactVerboseRegEx



def _isType(x) -> bool:
	return type is type(x)
#


_compileSingleType = None









class AbstractGenerator(object):

	#
	# @param	list _xParentalRequiredList				The JSON list that holds the names of those properties that are required. This list is defined at the parent!
	#
	def __init__(self, _xParent, _xName:str, _xSchema:dict, _xParentalRequiredList:list):
		self._schema = _xSchema
		self._parent = _xParent
		self._name = _xName
		self._requiredList = _xParentalRequiredList
		self._definitionID = None					# if this is a definition, this variable holds the name of the definition
	#

	#
	# Returns `true` if this is a definition. `false` indicates this is a regular schema component.
	#
	def isDefinition(self) -> bool:
		return bool(self._definitionID)
	#

	def required(self):
		if self._requiredList:
			if self._name not in self._requiredList:
				self._requiredList.append(self._name)
			return self
		else:
			raise Exception("This is not a value key that could be set to optional/required.")
	#

	def notRequired(self):
		if self._requiredList:
			if self._name in self._requiredList:
				self._requiredList.remove(self._name)
			return self
		else:
			raise Exception("This is not a value key that could be set to optional/required.")
	#

	def __str__(self):
		return jk_json.dumps(self._schema)
	#

	def __repr__(self):
		return jk_json.dumps(self._schema)
	#

	@property
	def schema(self):
		return self._schema
	#

	def __enter__(self):
		return self
	#

	def __exit__(self, etype, value, traceback):
		return False
	#

#



class BooleanGenerator(AbstractGenerator):

	def __init__(self, _xParent, _xName:str, _xSchema:dict, _xParentalRequiredList:list):
		super().__init__(_xParent, _xName, _xSchema, _xParentalRequiredList)
	#

#



class IntegerGenerator(AbstractGenerator):

	def __init__(self, _xParent, _xName:str, _xSchema:dict, _xParentalRequiredList:list):
		super().__init__(_xParent, _xName, _xSchema, _xParentalRequiredList)
	#

	def minimum(self, minimum:int):
		if minimum is None:
			if "minimum" in self._schema:
				self._schema.remove("minimum")
		else:
			self._schema["minimum"] = minimum
		return self
	#

	def exclusiveMinimum(self, minimum:int):
		if minimum is None:
			if "exclusiveMinimum" in self._schema:
				self._schema.remove("exclusiveMinimum")
		else:
			self._schema["exclusiveMinimum"] = minimum
		return self
	#

	def maximum(self, maximum:int):
		if maximum is None:
			if "maximum" in self._schema:
				self._schema.remove("maximum")
		else:
			self._schema["maximum"] = maximum
		return self
	#

	def exclusiveMaximum(self, maximum:int):
		if maximum is None:
			if "exclusiveMaximum" in self._schema:
				self._schema.remove("exclusiveMaximum")
		else:
			self._schema["exclusiveMaximum"] = maximum
		return self
	#

	def allowedValues(self, allowedValues:list):
		if allowedValues is None:
			if "enum" in self._schema:
				self._schema.remove("enum")
		else:
			self._schema["enum"] = allowedValues
		return self
	#

#



class FloatGenerator(AbstractGenerator):

	def __init__(self, _xParent, _xName:str, _xSchema:dict, _xParentalRequiredList:list):
		super().__init__(_xParent, _xName, _xSchema, _xParentalRequiredList)
	#

	def minimum(self, minimum:float):
		if minimum is None:
			if "minimum" in self._schema:
				self._schema.remove("minimum")
		else:
			self._schema["minimum"] = minimum
		return self
	#

	def exclusiveMinimum(self, minimum:float):
		if minimum is None:
			if "exclusiveMinimum" in self._schema:
				self._schema.remove("exclusiveMinimum")
		else:
			self._schema["exclusiveMinimum"] = minimum
		return self
	#

	def maximum(self, maximum:float):
		if maximum is None:
			if "maximum" in self._schema:
				self._schema.remove("maximum")
		else:
			self._schema["maximum"] = maximum
		return self
	#

	def exclusiveMaximum(self, maximum:float):
		if maximum is None:
			if "exclusiveMaximum" in self._schema:
				self._schema.remove("exclusiveMaximum")
		else:
			self._schema["exclusiveMaximum"] = maximum
		return self
	#

	def allowedValues(self, allowedValues:list):
		if allowedValues is None:
			if "enum" in self._schema:
				self._schema.remove("enum")
		else:
			self._schema["enum"] = allowedValues
		return self
	#

#



class StringGenerator(AbstractGenerator):

	def __init__(self, _xParent, _xName:str, _xSchema:dict, _xParentalRequiredList:list):
		super().__init__(_xParent, _xName, _xSchema, _xParentalRequiredList)
	#

	def minLength(self, minLength:int):
		if minLength is None:
			if "minLength" in self._schema:
				self._schema.remove("minLength")
		else:
			self._schema["minLength"] = minLength
		return self
	#

	def maxLength(self, maxLength:int):
		if maxLength is None:
			if "maxLength" in self._schema:
				self._schema.remove("maxLength")
		else:
			self._schema["maxLength"] = maxLength
		return self
	#

	def regexPattern(self, regexPattern:str):
		if regexPattern is None:
			if "enum" in self._schema:
				self._schema.remove("enum")
		else:
			if len(regexPattern) != len(regexPattern.strip()):
				raise Exception("Invalid pattern!")
			self._schema["pattern"] = regexPattern
		return self
	#

	def regexPatternVerbose(self, regexPattern:str):
		if regexPattern is None:
			if "enum" in self._schema:
				self._schema.remove("enum")
		else:
			regexPattern = compactVerboseRegEx(regexPattern)
			if len(regexPattern) != len(regexPattern.strip()):
				raise Exception("Invalid pattern!")
			self._schema["pattern"] = regexPattern
		return self
	#

	def allowedValues(self, allowedValues:list):
		if allowedValues is None:
			if "enum" in self._schema:
				self._schema.remove("enum")
		else:
			self._schema["enum"] = allowedValues
		return self
	#

#



class ListGenerator(AbstractGenerator):

	#
	# @param	AbtractGenerator[]			The generator or generators the list elements can be of
	#
	def __init__(self, _xParent, _xName:str, _xSchema:dict, _xParentalRequiredList:list, subGenList:list):
		super().__init__(_xParent, _xName, _xSchema, _xParentalRequiredList)

		assert isinstance(subGenList, (list, tuple))
		for subGen in subGenList:
			isinstance(subGen, AbstractGenerator)

		self.__subGenList = subGenList
	#

	@property
	def dataType(self) -> AbstractGenerator:
		if len(__subGenList) == 1:
			return self.__subGenList[0]
		else:
			raise Exception("There are multiple types defined!")
	#

	def minLength(self, minLength:int):
		if minLength is None:
			if "minItems" in self._schema:
				self._schema.remove("minItems")
		else:
			self._schema["minItems"] = minLength
		return self
	#

	def maxLength(self, maxLength:int):
		if maxLength is None:
			if "maxItems" in self._schema:
				self._schema.remove("maxItems")
		else:
			self._schema["maxItems"] = maxLength
		return self
	#

	def allowedValues(self, allowedValues:list):
		# TODO: The intention is to restrict the allowed values to certain values. Does this implementation here conform to the JSON schema specification?
		if allowedValues is None:
			if "enum" in self._schema:
				self._schema.remove("enum")
		else:
			self._schema["enum"] = allowedValues
		return self
	#

#



class ObjectGenerator(AbstractGenerator):

	def __init__(self, _xParent, _xName:str, _xSchema:dict, _xParentalRequiredList:list):
		super().__init__(_xParent, _xName, _xSchema, _xParentalRequiredList)
		if _xSchema is None:
			_xSchema = {
				"type": [ "object" ]
			}
		if not "properties" in _xSchema:
			_xSchema["properties"] = {}
		if not "required" in _xSchema:
			_xSchema["required"] = []
		self._schema = _xSchema
	#

	def objectValue(self, name:str, bRequired:bool = True):
		if bRequired:
			self._schema["required"].append(name)
		subSchema = {
			"type": [ "object" ]
		}
		self._schema["properties"][name] = subSchema
		return ObjectGenerator(self, None, subSchema, None)
	#

	def intValue(self, name:str, bRequired:bool = True) -> IntegerGenerator:
		if bRequired:
			self._schema["required"].append(name)
		subSchema = {
			"type": [ "integer" ]
		}
		self._schema["properties"][name] = subSchema
		return IntegerGenerator(self, name, subSchema, self._schema["required"])
	#

	def floatValue(self, name:str, bRequired:bool = True) -> FloatGenerator:
		if bRequired:
			self._schema["required"].append(name)
		subSchema = {
			"type": [ "number" ]
		}
		self._schema["properties"][name] = subSchema
		return FloatGenerator(self, name, subSchema, self._schema["required"])
	#

	def boolValue(self, name:str, bRequired:bool = True) -> BooleanGenerator:
		if bRequired:
			self._schema["required"].append(name)
		subSchema = {
			"type": [ "boolean" ]
		}
		self._schema["properties"][name] = subSchema
		return BooleanGenerator(self, name, subSchema, self._schema["required"])
	#

	def strValue(self, name:str, bRequired:bool = True) -> StringGenerator:
		if bRequired:
			self._schema["required"].append(name)
		subSchema = {
			"type": [ "string" ]
		}
		self._schema["properties"][name] = subSchema
		return StringGenerator(self, name, subSchema, self._schema["required"])
	#

	#
	# A property should be of type "array".
	#
	# @param		str name							The name of the property.
	# @param		type|AbstractGenerator listType		The type of the property values. (All property values must be of exactly this single type specified here.)
	# @param		bool bRequired						This property is either optional or required.
	#
	def listValue(self, name:str, listType:typing.Union[type,AbstractGenerator], bRequired:bool = True) -> ListGenerator:
		if bRequired:
			self._schema["required"].append(name)

		# ----

		subTypeSchema, subGens = _compileListType(self, listType)

		# ----

		subSchema = {
			"type": [ "array" ],
			"items": subTypeSchema,
		}
		self._schema["properties"][name] = subSchema
		return ListGenerator(self, name, subSchema, self._schema["required"], subGens)
	#

#









_SUB_TYPE_CLASSES_BY_TYPE = {
	bool: (BooleanGenerator, "boolean"),
	int: (IntegerGenerator, "integer"),
	float: (FloatGenerator, "number"),
	str: (StringGenerator, "string"),
}

_SUB_TYPE_CLASSES_BY_STR = {
	"bool": (BooleanGenerator, "boolean"),
	"boolean": (BooleanGenerator, "boolean"),
	"int": (IntegerGenerator, "integer"),
	"integer": (IntegerGenerator, "integer"),
	"float": (FloatGenerator, "number"),
	"number": (FloatGenerator, "number"),
	"str": (StringGenerator, "string"),
	"string": (StringGenerator, "string"),
}




def _compileSingleType(parent, listType:typing.Union[type,str,"AbstractGenerator"]) -> tuple:
	if isinstance(listType, AbstractGenerator):
		if not listType.isDefinition:
			raise Exception("The specified list element type is not a definition!")

		subTypeSchema = dict(listType.schema)
		subGen = listType

	elif isinstance(listType, str):
		if listType in _SUB_TYPE_CLASSES_BY_STR:
			genClazz, jsType = _SUB_TYPE_CLASSES_BY_STR[listType]
			subTypeSchema = {
				"type": jsType
			}
			subGen = genClazz(parent, None, subTypeSchema, None)
		else:
			raise Exception("Invalid list element type specified: " + str(listType))

	elif _isType(listType):
		if listType in _SUB_TYPE_CLASSES_BY_TYPE:
			genClazz, jsType = _SUB_TYPE_CLASSES_BY_TYPE[listType]
			subTypeSchema = {
				"type": jsType
			}
			subGen = genClazz(parent, None, subTypeSchema, None)
		else:
			raise Exception("Invalid list element type specified: " + str(listType))

	else:
		raise Exception("Invalid list element type specified: " + str(listType))

	return subTypeSchema, subGen
#

def _compileListType(parent, listType:typing.Union[type,"AbstractGenerator"]) -> tuple:
	subTypeSchema, subGen = _compileSingleType(parent, listType)
	return subTypeSchema, [ subGen ]
#











class _Generator(object):

	def __init__(self):
		self.__schema = None
		self.__rng = random.Random()
		self.__defs = {}
	#

	def __generateID(self, existingIDs):
		CHARS = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"

		while True:
			randID = "".join([ self.__rng.choice(CHARS) for i in range(0, 8) ])
			if randID not in existingIDs:
				return randID
	#

	def __str__(self):
		return jk_json.dumps(self.__schema)
	#

	def __repr__(self):
		return jk_json.dumps(self.__schema)
	#

	def objectValue(self) -> ObjectGenerator:
		if self.__schema is not None:
			raise Exception("This generator already provides a schema!")

		# ----

		ret = ObjectGenerator(None, None, None, None)
		self.__schema = ret._schema
		self.__schema["$schema"] = "http://json-schema.org/draft-04/schema#"
		return ret
	#

	#
	# A property should be of type "array".
	#
	# @param		str name							The name of the property.
	# @param		type|AbstractGenerator listType		The type of the property values. (All property values must be of exactly this single type specified here.)
	# @param		bool bRequired						This property is either optional or required.
	#
	def listValue(self, listType:typing.Union[type,AbstractGenerator], bRequired:bool = True) -> ListGenerator:
		if self.__schema is not None:
			raise Exception("This generator already provides a schema!")

		# ----

		subTypeSchema, subGens = _compileListType(self, listType)

		# ----

		subSchema = {
			"type": [ "array" ],
			"items": subTypeSchema,
		}
		ret = ListGenerator(self, None, subSchema, None, subGens)
		self.__schema = ret._schema
		self.__schema["$schema"] = "http://json-schema.org/draft-04/schema#"
		return ret
	#

	@property
	def schema(self):
		ret = dict(self.__schema)
		return ret
	#

	def __enter__(self):
		return self
	#

	def __exit__(self, etype, value, traceback):
		return False
	#

	#
	# Invoke this method to define an object schema that can be used as a component in other definitions later.
	#
	def defineObject(self, name:str = None) -> ObjectGenerator:
		if name is None:
			name = self.__generateID(list(self.__defs.keys()))
		else:
			assert isinstance(name, str)
			assert re.match("^[a-zA-Z]+$", name)
			assert name not in self.__defs

		ret = ObjectGenerator(None, None, None, None)
		ret._definitionID = name
		self.__defs[name] = ret
		return ret
	#

#











def createObjectSchemaGenerator() -> ObjectGenerator:
	return ObjectGenerator(None, None, None, None)
#

def createSchemaGenerator():
	return _Generator()
#









#!/usr/bin/env python3
# -*- coding: utf-8 -*-








class _SchemaAST(object):

	def __init__(self,
		any_allowedTypes,									# list<int>
		any_allowedValues,									# list<*>
		any_mustBeExactValue,								# 1tuple<*>
		number_multipleValueOf,								# int|float
		number_maximum,										# int|float
		number_exclusiveMaximum,							# int|float
		number_minimum,										# int|float
		number_exclusiveMinimum,							# int|float
		string_minLength,									# int
		string_maxLength,									# int
		string_pattern,										# regex
		array_items,										# _SchemaAST|_SchemaAST[]
		array_additionalItems,								# _SchemaAST|_SchemaAST[]
		array_maxItems,										# int
		array_minItems,										# int
		array_itemsMustBeUnique,							# bool
		array_mustContainExactValue,						# _SchemaAST
		array_itemsMustNotBeNone,							# bool
		array_itemsMustBeNone,								# bool
		object_minProperties,								# int
		object_maxProperties,								# int
		object_properties,									# dict<str,_SchemaAST>
		object_patternProperties,							# list<(str,_SchemaAST)>
		object_propertyDependencyObjectMustMatchSchema,		# dict<str,_SchemaAST>
		object_propertyDependencyOtherPropertyMustExist,	# dict<str,str[]>
		object_propertyNames,								# _SchemaAST
		object_requiredProperties,							# str[]
		object_additionalProperties,						# _SchemaAST
		op_if,												# _SchemaAST
		op_then,											# _SchemaAST
		op_else,											# _SchemaAST
		op_not,												# _SchemaAST
		op_allOf,											# list<_SchemaAST>
		op_anyOf,											# list<_SchemaAST>
		op_oneOf,											# list<_SchemaAST>
		always = None										# bool
	):
		self.any_allowedTypes = any_allowedTypes
		self.any_allowedValues = any_allowedValues
		self.any_mustBeExactValue = any_mustBeExactValue
		self.number_multipleValueOf = number_multipleValueOf
		self.number_maximum = number_maximum
		self.number_exclusiveMaximum = number_exclusiveMaximum
		self.number_minimum = number_minimum
		self.number_exclusiveMinimum = number_exclusiveMinimum
		self.string_minLength = string_minLength
		self.string_maxLength = string_maxLength
		self.string_pattern = string_pattern
		self.array_items = array_items
		self.array_additionalItems = array_additionalItems
		self.array_maxItems = array_maxItems
		self.array_minItems = array_minItems
		self.array_itemsMustBeUnique = array_itemsMustBeUnique
		self.array_mustContainExactValue = array_mustContainExactValue
		self.array_itemsMustNotBeNone = array_itemsMustNotBeNone
		self.array_itemsMustBeNone = array_itemsMustBeNone
		self.object_minProperties = object_minProperties
		self.object_maxProperties = object_maxProperties
		self.object_properties = object_properties
		self.object_patternProperties = object_patternProperties
		self.object_propertyDependencyObjectMustMatchSchema = object_propertyDependencyObjectMustMatchSchema
		self.object_propertyDependencyOtherPropertyMustExist = object_propertyDependencyOtherPropertyMustExist
		self.object_propertyNames = object_propertyNames
		self.object_requiredProperties = object_requiredProperties
		self.object_additionalProperties = object_additionalProperties
		self.op_if = op_if
		self.op_then = op_then
		self.op_else = op_else
		self.op_not = op_not
		self.op_allOf = op_allOf
		self.op_anyOf = op_anyOf
		self.op_oneOf = op_oneOf
		self.always = always
	#

	def __makeNullIfEmptyListOrDict(self, item):
		if item is None:
			return None

		if len(item) == 0:
			return None
		else:
			return item
	#

	def __normalize_makeNullIfEmptyListOrDict(self, item):
		if item is None:
			return None

		if isinstance(item, list):
			for f in item:
				f.normalize()
		elif isinstance(item, dict):
			newItem = {}
			for k, f in item.items():
				if f != None:
					f.normalize()
					newItem[k] = f
			item = newItem

		if len(item) == 0:
			return None
		else:
			return item
	#

	def __normalize_makeNullIfEmptyAST(self, ast):
		if ast is None:
			return None

		ast.normalize()
		if ast.isNotEmpty():
			return ast
		else:
			return None
	#

	def normalize(self):
		self.any_allowedTypes = self.__makeNullIfEmptyListOrDict(self.any_allowedTypes)
		self.any_allowedValues = self.__makeNullIfEmptyListOrDict(self.any_allowedValues)

		newList = []
		for regex, schema in self.object_patternProperties:
			schema = self.__normalize_makeNullIfEmptyAST(schema)
			if schema != None:
				newList.append((regex, schema))
		if len(newList) == 0:
			newList = None
		self.object_patternProperties = newList

		self.array_items = self.__normalize_makeNullIfEmptyListOrDict(self.array_items)
		self.object_properties = self.__normalize_makeNullIfEmptyListOrDict(self.object_properties)
		self.object_propertyDependencyObjectMustMatchSchema = self.__normalize_makeNullIfEmptyListOrDict(self.object_propertyDependencyObjectMustMatchSchema)
		self.op_allOf = self.__normalize_makeNullIfEmptyListOrDict(self.op_allOf)
		self.op_anyOf = self.__normalize_makeNullIfEmptyListOrDict(self.op_anyOf)
		self.op_oneOf = self.__normalize_makeNullIfEmptyListOrDict(self.op_oneOf)

		newDict = {}
		for key, stringList in self.object_propertyDependencyOtherPropertyMustExist.items():
			if len(stringList) > 0:
				newDict[key] = stringList
		if len(newDict) == 0:
			newDict = None
		self.object_propertyDependencyOtherPropertyMustExist[key] = newDict

		self.array_additionalItems = self.__normalize_makeNullIfEmptyAST(self.array_additionalItems)
		self.array_mustContainExactValue = self.__normalize_makeNullIfEmptyAST(self.array_mustContainExactValue)
		self.object_propertyNames = self.__normalize_makeNullIfEmptyAST(self.object_propertyNames)
		self.op_if = self.__normalize_makeNullIfEmptyAST(self.op_if)
		self.op_then = self.__normalize_makeNullIfEmptyAST(self.op_then)
		self.op_else = self.__normalize_makeNullIfEmptyAST(self.op_else)
		self.op_not = self.__normalize_makeNullIfEmptyAST(self.op_not)
	#

	def isNotEmpty(self):
		return									\
				self.hasAnyTypeConstraint()		\
				or self.hasStringConstraint() 	\
				or self.hasArrayConstraint()	\
				or self.hasObjectConstraint()	\
				or self.hasNumberConstraint()	\
				or self.hasOpConstraint()
	#

	def isEmpty(self):
		return not								\
			(									\
				self.hasAnyTypeConstraint()		\
				or self.hasStringConstraint()	 \
				or self.hasArrayConstraint()	\
				or self.hasObjectConstraint()	\
				or self.hasNumberConstraint()	\
				or self.hasOpConstraint()		\
			)
	#

	def hasAnyTypeConstraint(self):
		return (self.any_allowedTypes != None)			\
		 	or (self.any_allowedValues != None)			\
		 	or (self.any_mustBeExactValue != None)		\
		 	or (self.always != None)
	#

	def hasStringConstraint(self):
		return (self.string_minLength != None)			\
		 	or (self.string_maxLength != None)			\
		 	or (self.string_pattern != None)
	#

	def hasArrayConstraint(self):
		return (self.array_items != None)				\
		 	or (self.array_additionalItems != None)		\
		 	or (self.array_maxItems != None)			\
		 	or (self.array_minItems != None)			\
		 	or (self.array_itemsMustBeUnique != None)	\
		 	or (self.array_mustContainExactValue != None)
	#

	def hasObjectConstraint(self):
		return (self.object_minProperties != None)					\
			or (self.object_maxProperties != None)					\
			or (self.object_properties != None)						\
		 	or (self.object_patternProperties != None)				\
		 	or (self.object_propertyDependencyOtherPropertyMustExist != None)	\
		 	or (self.object_propertyDependencyOtherPropertyMustExist != None)	\
		 	or (self.object_propertyNames != None)					\
			or (self.object_requiredProperties != None)
	#

	def hasOpConstraint(self):
		return (self.op_if != None)						\
		 	or (self.op_not != None)					\
		 	or (self.op_allOf != None)					\
		 	or (self.op_anyOf != None)					\
		 	or (self.op_oneOf != None)
	#

	def hasNumberConstraint(self):
		return (self.number_multipleValueOf != None)	\
		 	or (self.number_maximum != None)			\
		 	or (self.number_exclusiveMaximum != None)	\
		 	or (self.number_minimum != None)			\
		 	or (self.number_exclusiveMinimum != None)	\
	#

#





ALWAYS_TRUE = _SchemaAST(
	None, None, None, None, None, None, None, None, None, None,
	None, None, None, None, None, None, None, None, None, None,
	None, None, None, None, None, None, None, None, None, None,
	None, None, None, None, None,
	True)

ALWAYS_FALSE = _SchemaAST(
	None, None, None, None, None, None, None, None, None, None,
	None, None, None, None, None, None, None, None, None, None,
	None, None, None, None, None, None, None, None, None, None,
	None, None, None, None, None,
	False)























__version__ = "0.2021.1.18"



import gzip
import json as __json
import bz2

import chardet

from .ObjectEncoder import ObjectEncoder

from .Token import Token
from .TokenStream import TokenStream
from .TokenizerBase import TokenizerBase
from .SourceCodeLocation import SourceCodeLocation
from .ParserErrorException import ParserErrorException

from .TokenizerRelaxed import TokenizerRelaxed
from .TokenizerStrict import TokenizerStrict
from .JsonParserRelaxed import JsonParserRelaxed
from .JsonParserStrict import JsonParserStrict




__parserRelaxed = JsonParserRelaxed()
__tokenizerRelaxed = TokenizerRelaxed()

__parserStrict = JsonParserStrict()
__tokenizerStrict = TokenizerStrict()



#
# Deserialize a JSON string: Reconstruct a python data structure from the specified JSON string.
#
# @param	str textToParse		The JSON to parse in binary or string representation. If binary data is specified a simple UTF-8 decoding is performed
#								to get to a string which is then parsed.
# @param	bool bStrict		If ```True``` this parser sticks strictly to the JSON standard. If ```False``` C-style comments
#								are allowed and strings can be specified with single quotes and double quotes.
#								Furthermore NaN, positive and negative infinitiy is supported.
#
def loads(textToParse:str, bStrict:bool = False, bDebugging:bool = False):
	assert isinstance(textToParse, str)
	assert isinstance(bStrict, bool)
	assert isinstance(bDebugging, bool)

	if isinstance(textToParse, (bytes, bytearray)):
		textToParse = textToParse.decode("utf-8")
	elif not isinstance(textToParse, str):
		raise Exception("Can't decode JSON data from a non-string or non-binary value!")

	if bStrict:
		tokenizer = __tokenizerStrict
		parser = __parserStrict
	else:
		tokenizer = __tokenizerRelaxed
		parser = __parserRelaxed

	return parser.parse(tokenizer.tokenize(textToParse), bDebugging)
#



#
# Deserialize a JSON string: Reconstruct a python data structure reading data from the specified JSON file.
#
# @param	File fp				A file like object. All data contained in that file is read.
# @param	bool bStrict		If ```True``` this parser sticks strictly to the JSON standard. If ```False``` C-style comments
#								are allowed and strings can be specified with single quotes and double quotes.
#								Furthermore NaN, positive and negative infinitiy is supported.
#
def load(fp, bStrict = False, bDebugging = False):
	rawData = fp.read()
	return loads(rawData, bStrict = bStrict, bDebugging = bDebugging)
#



#
# Deserialize a JSON string: Reconstruct a python data structure reading data from the specified JSON file.
#
# @param	str filePath		The path of the file to load.
# @param	bool bStrict		If ```True``` this parser sticks strictly to the JSON standard. If ```False``` C-style comments
#								are allowed and strings can be specified with single quotes and double quotes.
#								Furthermore NaN, positive and negative infinitiy is supported.
#
def loadFromFile(filePath:str, bStrict:bool = False, bDebugging:bool = False, encoding:str = None, autoDetectEncoding:bool = True):
	assert isinstance(filePath, str)
	assert isinstance(bStrict, bool)
	assert isinstance(bDebugging, bool)

	if filePath.endswith(".bz2"):
		with bz2.open(filePath, "rb") as f:
			rawData = f.read()
	elif filePath.endswith(".gz"):
		with gzip.open(filePath, "rb") as f:
			rawData = f.read()
	else:
		with open(filePath, "rb") as f:
			rawData = f.read()

	textData = None

	if encoding is None:
		if autoDetectEncoding:
			try:
				if rawData.startswith(b"\xef\xbb\xbf"):		# utf-8 byte order mark
					rawData = rawData[3:]

				textData = rawData.decode("utf-8")

			except:
				encoding = chardet.detect(rawData)["encoding"]
				if encoding is None:
					encoding = "utf-8"

		else:
			encoding = "utf-8"

	if textData is None:
		textData = rawData.decode(encoding)

	return loads(textData, bStrict = bStrict, bDebugging = bDebugging)
#



#
# Serialize the specified object or value to a JSON formatted JSON string.
#
# @param		mixed jsonObj		The object to serialize
# @param		str indent			The indentation to use
# @param		bool sort_keys		It ```True``` sort the properties
#
def dumps(jsonObj, indent=None, sort_keys=False, linePrefix=None, cls=None):
	assert isinstance(jsonObj, (str, int, float, bool, tuple, list, dict, type(None)))

	# for now we rely on the default json serializer/deserializer
	if indent is None:
		return __json.dumps(jsonObj, indent=None, separators=(',', ':'), sort_keys=sort_keys, cls=cls)
	else:
		if linePrefix is None:
			return __json.dumps(jsonObj, indent=indent, sort_keys=sort_keys, cls=cls)
		else:
			lines = __json.dumps(jsonObj, indent=indent, sort_keys=sort_keys, cls=cls).split("\n")
			ret = linePrefix + ("\n" + linePrefix).join(lines) + "\n"
			return ret
#



def dump(jsonObj, f, indent=None, sort_keys=False, linePrefix=None, cls=None):
	assert isinstance(jsonObj, (str, int, float, bool, tuple, list, dict, type(None)))

	# for now we rely on the default json serializer/deserializer
	if indent is None:
		return __json.dump(jsonObj, f, indent=None, separators=(',', ':'), sort_keys=sort_keys, cls=cls)
	else:
		if linePrefix is None:
			return __json.dump(jsonObj, f, indent=indent, sort_keys=sort_keys, cls=cls)
		else:
			lines = __json.dump(jsonObj, f, indent=indent, sort_keys=sort_keys, cls=cls).split("\n")
			ret = linePrefix + lines.join("\n" + linePrefix) + "\n"
			return ret
#



def saveToFile(jsonObj, filePath, indent=None, sort_keys=False, linePrefix=None):
	assert isinstance(jsonObj, (str, int, float, bool, tuple, list, dict, type(None)))

	with open(filePath, "w", encoding="utf-8") as f:
		dump(jsonObj, f, indent=indent, sort_keys=sort_keys, linePrefix=linePrefix)
		#__json.dump(jsonObj, f, indent=indent, sort_keys=sort_keys)
#



def saveToFilePretty(jsonObj, filePath:str, linePrefix=None):
	assert isinstance(jsonObj, (str, int, float, bool, tuple, list, dict, type(None)))
	assert isinstance(filePath, str)
	if linePrefix is not None:
		assert isinstance(linePrefix, str)

	with open(filePath, "w", encoding="utf-8") as f:
		dump(jsonObj, f, indent="\t", sort_keys=True, linePrefix=linePrefix, cls=ObjectEncoder)
		#__json.dump(jsonObj, f, indent=indent, sort_keys=sort_keys)
#



def prettyPrint(jsonObj, linePrefix=None):
	assert isinstance(jsonObj, (str, int, float, bool, tuple, list, dict, type(None)))

	print(dumps(jsonObj, indent="\t", sort_keys=True, linePrefix=linePrefix, cls=ObjectEncoder))
#


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import math

from .TokenStream import *
from .ParserErrorException import *
from .ParserBase import *




class JsonParserStrict(ParserBase):

	def __init__(self):
		pass
	#

	#
	# Parse the tokens. Construct an abstract syntax tree from the tokens provided by the tokenizer.
	#
	# @param	Tokenizer tokenizer		The tokenizer that provides the tokens.
	#
	def parse(self, tokenizer, bDebugging = False):
		if isinstance(tokenizer, TokenStream):
			ts = tokenizer
		else:
			ts = TokenStream(tokenizer)

		ctx = ParsingContext(bDebugging)
		bSuccess, data = self._tryEat_S(ctx, ts)
		if bSuccess:
			return data
		else:
			raise Exception()
	#

	@debugDecorator
	def _tryEat_S(self, ctx, ts):
		# loc = ts.location

		n = ts.skipAll("eol", None)

		bSuccess, parsedData = self._tryEat_JANYVALUE(ctx, ts)
		if not bSuccess:
			raise ParserErrorException(ts.location, "Syntax error: Expecting valid JSON value element")

		n = ts.skipAll("eol", None)

		#if not ts.isEOS:
		#	raise ParserErrorException(ts.location, "Syntax error")

		return (True, parsedData)
	#

	@debugDecorator
	def _tryEat_JANYVALUE(self, ctx, ts):
		# loc = ts.location

		t1 = ts.peek()
		if t1.type == "w":
			if t1.text == "null":
				ts.skip()
				return (True, None)
			if t1.text == "true":
				ts.skip()
				return (True, True)
			if t1.text == "false":
				ts.skip()
				return (True, False)
			if t1.text == "Infinity":
				ts.skip()
				return (True, math.inf)
			if t1.text == "NaN":
				ts.skip()
				return (True, math.nan)
			raise ParserErrorException(ts.location, "Syntax error: Expecting valid JSON value element")
		if t1.type == "s":
			ts.skip()
			return (True, t1.text)
		if t1.type == "i":
			ts.skip()
			return (True, int(t1.text))
		if t1.type == "f":
			ts.skip()
			return (True, float(t1.text))

		bSuccess, parsedData = self._tryEat_JARRAY(ctx, ts)
		if bSuccess:
			return (True, parsedData)

		bSuccess, parsedData = self._tryEat_JOBJECT(ctx, ts)
		if bSuccess:
			return (True, parsedData)

		if (t1.type == "d") and (t1.text == "-"):
			m = ts.mark()
			ts.skip()
			t2 = ts.peek()
			if (t2.type == "w") and (t2.text == "Infinity"):
				ts.skip()
				return (True, float("-inf"))
			m.resetToMark()

		return (False, None)
	#

	@debugDecorator
	def _tryEat_JARRAY(self, ctx, ts):
		# loc = ts.location

		t1 = ts.peek()
		if (t1.type != "d") or (t1.text != "["):
			return (False, None)
		ts.skip()

		bSuccess, parsedData = self._tryEat_JVALUE_LIST(ctx, ts)
		if not bSuccess:
			parsedData = []

		t1 = ts.peek()
		if (t1.type != "d") or (t1.text != "]"):
			raise ParserErrorException(ts.location, "Syntax error: Expecting JSON array elements or ']'")
		ts.skip()

		return (True, parsedData)
	#

	@debugDecorator
	def _tryEat_JVALUE_LIST(self, ctx, ts):
		# loc = ts.location

		retData = []
		bSuccess, parsedData = self._tryEat_JANYVALUE(ctx, ts)
		if not bSuccess:
			return (False, None)
		retData.append(parsedData)

		while True:
			t1 = ts.peek()
			if (t1.type != "d") or (t1.text != ","):
				return (True, retData)
			ts.skip()

			bSuccess, parsedData = self._tryEat_JANYVALUE(ctx, ts)
			if not bSuccess:
				raise ParserErrorException(ts.location, "Syntax error: Expecting valid JSON value element")
			retData.append(parsedData)
	#

	@debugDecorator
	def _tryEat_JOBJECT(self, ctx, ts):
		# loc = ts.location

		t1 = ts.peek()
		if (t1.type != "d") or (t1.text != "{"):
			return (False, None)
		ts.skip()

		bSuccess, parsedData = self._tryEat_JPROPERTY_LIST(ctx, ts)
		if not bSuccess:
			parsedData = {}

		t1 = ts.peek()
		if (t1.type != "d") or (t1.text != "}"):
			raise ParserErrorException(ts.location, "Syntax error: Expecting JSON property or '}'")
		ts.skip()

		return (True, parsedData)
	#

	@debugDecorator
	def _tryEat_JPROPERTY_LIST(self, ctx, ts):
		# loc = ts.location

		retData = {}
		bSuccess, parsedData = self._tryEat_JPROPERTY(ctx, ts)
		if not bSuccess:
			return (False, None)

		if parsedData[0] in retData:
			raise ParserErrorException(ts.location, "Syntax error: Duplicate property name detected: " + repr(parsedData[0]))
		retData[parsedData[0]] = parsedData[1]

		while True:
			t1 = ts.peek()
			if (t1.type != "d") or (t1.text != ","):
				return (True, retData)
			ts.skip()

			bSuccess, parsedData = self._tryEat_JPROPERTY(ctx, ts)
			if not bSuccess:
				raise ParserErrorException(ts.location, "Syntax error: Expecting JSON property")

			if parsedData[0] in retData:
				raise ParserErrorException(ts.location, "Syntax error: Duplicate property key detected: " + repr(parsedData[0]))
			retData[parsedData[0]] = parsedData[1]
	#

	@debugDecorator
	def _tryEat_JPROPERTY(self, ctx, ts):
		# loc = ts.location

		t1 = ts.peek()
		if t1.type != "s":
			return (False, None)
		ts.skip()

		t2 = ts.peek()
		if (t2.type != "d") or (t2.text != ":"):
			raise ParserErrorException(ts.location, "Syntax error: Expecting ':' followed by a property value!")
		ts.skip()

		bSuccess, parsedData = self._tryEat_JANYVALUE(ctx, ts)
		if bSuccess:
			return (True, (t1.text, parsedData))

		raise ParserErrorException(ts.location, "Syntax error: Expecting a valid JSON value!")
	#

#













# ================================================================================================================================
# ================================================================================================================================
# ================================================================================================================================



TYPE_NULL = 0
TYPE_BOOL = 1
TYPE_INT = 2
TYPE_FLOAT = 3
TYPE_STR = 4
TYPE_ARRAY = 5
TYPE_OBJECT= 6

ALL_TYPES_LIST = [ TYPE_NULL, TYPE_BOOL, TYPE_INT, TYPE_FLOAT, TYPE_STR, TYPE_ARRAY, TYPE_OBJECT ]
ALL_TYPES_NAME_LIST = [ "null", "boolean", "integer", "float", "string", "array", "object" ]
ALL_TYPES_SET = frozenset(ALL_TYPES_LIST)



# ================================================================================================================================
# ================================================================================================================================
# ================================================================================================================================



def getTypeID(typeName:str) -> int:
	if typeName == "null":
		return TYPE_NULL
	if typeName == "boolean":
		return TYPE_BOOL
	if typeName == "integer":
		return TYPE_INT
	if typeName == "number":
		return TYPE_FLOAT
	if typeName == "string":
		return TYPE_STR
	if typeName == "array":
		return TYPE_ARRAY
	if typeName == "object":
		return TYPE_OBJECT
	raise Exception("Invalid type: " + str(typeName))
#

def getTypeIDs(typeNames:list):
	ret = []
	for typeName in typeNames:
		ret.append(getTypeID(typeName))
	return ret
#

def getTypeIDOfType(pythonType:type) -> int:
	if (pythonType is None) or (pythonType is type(None)):
		return TYPE_NULL
	if pythonType == bool:
		return TYPE_BOOL
	if pythonType == int:
		return TYPE_INT
	if pythonType == float:
		return TYPE_FLOAT
	if pythonType == str:
		return TYPE_STR
	if (pythonType == list) or (pythonType == tuple):
		return TYPE_ARRAY
	if pythonType == dict:
		return TYPE_OBJECT
	raise Exception("Invalid type: " + str(pythonType))
#

def getTypeIDOfTypeNonInt(pythonType:type) -> int:
	if (pythonType is None) or (pythonType is type(None)):
		return TYPE_NULL
	if pythonType == bool:
		return TYPE_BOOL
	if (pythonType == int) or (pythonType == float):
		return TYPE_FLOAT
	if pythonType == str:
		return TYPE_STR
	if (pythonType == list) or (pythonType == tuple):
		return TYPE_ARRAY
	if pythonType == object:
		return TYPE_OBJECT
	raise Exception("Invalid type: " + str(pythonType))
#

def getTypeIDOfValue(value) -> int:
	if value is None:
		return TYPE_NULL
	if isinstance(value, bool):
		return TYPE_BOOL
	if isinstance(value, int):
		return TYPE_INT
	if isinstance(value, float):
		return TYPE_FLOAT
	if isinstance(value, str):
		return TYPE_STR
	if isinstance(value, (list, tuple)):
		return TYPE_ARRAY
	if isinstance(value, dict):
		return TYPE_OBJECT
	raise Exception("Invalid type: " + str(type(value)))
#

def getTypeIDOfValueNonInt(value) -> int:
	if value is None:
		return TYPE_NULL
	if isinstance(value, bool):
		return TYPE_BOOL
	if isinstance(value, (int, float)):
		return TYPE_FLOAT
	if isinstance(value, str):
		return TYPE_STR
	if isinstance(value, (list, tuple)):
		return TYPE_ARRAY
	if isinstance(value, dict):
		return TYPE_OBJECT
	raise Exception("Invalid type: " + str(type(value)))
#


# ================================================================================================================================
# ================================================================================================================================
# ================================================================================================================================



def _cmp_list(jsonListA, jsonListB):
	if len(jsonListA) != len(jsonListB):
		return False

	for va, vb in zip(jsonListA, jsonListB):
		ta = getTypeIDOfValue(va)
		tb = getTypeIDOfValue(vb)

		if ta != tb:
			return False

		if ta == TYPE_ARRAY:
			return _cmp_list(va, vb)
		elif ta == TYPE_OBJECT:
			return _cmp_obj(va, vb)
		else:
			return va == vb

	return True
#

def _cmp_list_nonInt(jsonListA, jsonListB):
	if len(jsonListA) != len(jsonListB):
		return False

	for va, vb in zip(jsonListA, jsonListB):
		ta = getTypeIDOfValueNonInt(va)
		tb = getTypeIDOfValueNonInt(vb)

		if ta != tb:
			return False

		if ta == TYPE_ARRAY:
			return _cmp_list_nonInt(va, vb)
		elif ta == TYPE_OBJECT:
			return _cmp_obj_nonInt(va, vb)
		else:
			return va == vb

	return True
#

def _cmp_obj(jsonObjA, jsonObjB):
	if len(jsonObjA) != len(jsonObjB):
		return False

	for propertyName in jsonObjA:
		if propertyName not in jsonObjB:
			return False

		va = jsonObjA[propertyName]
		vb = jsonObjB[propertyName]

		ta = getTypeIDOfValue(va)
		tb = getTypeIDOfValue(vb)
		if ta != tb:
			return False

		if ta == TYPE_ARRAY:
			return _cmp_list(va, vb)
		elif ta == TYPE_OBJECT:
			return _cmp_obj(va, vb)
		else:
			return va == vb

	return True
#

def _cmp_obj_nonInt(jsonObjA, jsonObjB):
	if len(jsonObjA) != len(jsonObjB):
		return False

	for propertyName in jsonObjA:
		if propertyName not in jsonObjB:
			return False

		va = jsonObjA[propertyName]
		vb = jsonObjB[propertyName]

		ta = getTypeIDOfValueNonInt(va)
		tb = getTypeIDOfValueNonInt(vb)
		if ta != tb:
			return False

		if ta == TYPE_ARRAY:
			return _cmp_list_nonInt(va, vb)
		elif ta == TYPE_OBJECT:
			return _cmp_obj_nonInt(va, vb)
		else:
			return va == vb

	return True
#

def jsonIsEqual(va, vb, dontDistinguishBetweenIntAndFloat = True):
	if dontDistinguishBetweenIntAndFloat:
		ta = getTypeIDOfValueNonInt(va)
		tb = getTypeIDOfValueNonInt(vb)
		# print(str(ta) + "\t" + repr(va))
		# print(str(tb) + "\t" + repr(vb))

		if ta != tb:
			return False

		if ta == TYPE_ARRAY:
			return _cmp_list_nonInt(va, vb)
		elif ta == TYPE_OBJECT:
			return _cmp_obj_nonInt(va, vb)
		else:
			return va == vb

	else:
		ta = getTypeIDOfValue(va)
		tb = getTypeIDOfValue(vb)
		# print(str(ta) + "\t" + repr(va))
		# print(str(tb) + "\t" + repr(vb))

		if ta != tb:
			return False

		if ta == TYPE_ARRAY:
			return _cmp_list(va, vb)
		elif ta == TYPE_OBJECT:
			return _cmp_obj(va, vb)
		else:
			return va == vb
#










#!/usr/bin/env python3
# -*- coding: utf-8 -*-




import collections





Token = collections.namedtuple("Token", [ "type", "text", "sourceID", "lineNo", "charPos", "endLineNo", "endCharPos" ])

















#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import math

from .TokenStream import *
from .ParserErrorException import *
from .ParserBase import *
from .JsonParserStrict import *



class JsonParserRelaxed(JsonParserStrict):

	def __init__(self):
		pass
	#

	@debugDecorator
	def _tryEat_JVALUE_LIST(self, ctx, ts):
		# loc = ts.location

		retData = []
		bSuccess, parsedData = self._tryEat_JANYVALUE(ctx, ts)
		if not bSuccess:
			return (False, None)
		retData.append(parsedData)

		while True:
			t1 = ts.peek()
			if (t1.type != "d") or (t1.text != ","):
				return (True, retData)
			ts.skip()

			bSuccess, parsedData = self._tryEat_JANYVALUE(ctx, ts)
			if not bSuccess:
				return (True, retData)
			retData.append(parsedData)
	#

	@debugDecorator
	def _tryEat_JPROPERTY_LIST(self, ctx, ts):
		# loc = ts.location

		retData = {}
		bSuccess, parsedData = self._tryEat_JPROPERTY(ctx, ts)
		if not bSuccess:
			return (False, None)

		if parsedData[0] in retData:
			raise ParserErrorException(ts.location, "Syntax error: Duplicate property name detected: " + repr(parsedData[0]))
		retData[parsedData[0]] = parsedData[1]

		while True:
			t1 = ts.peek()
			if (t1.type != "d") or (t1.text != ","):
				return (True, retData)
			ts.skip()

			bSuccess, parsedData = self._tryEat_JPROPERTY(ctx, ts)
			if not bSuccess:
				return (True, retData)

			if parsedData[0] in retData:
				raise ParserErrorException(ts.location, "Syntax error: Duplicate property key detected: " + repr(parsedData[0]))
			retData[parsedData[0]] = parsedData[1]
	#

#









#!/usr/bin/python3


import json
import os
import sys
import re

import jk_console
import jk_sysinfo
import jk_json
import jk_flexdata
import jk_logging
from jk_typing import *
import jk_argparsing









ap = jk_argparsing.ArgsParser("jsonPrettyPrint", "Format JSON data.")

ap.optionDataDefaults.set("help", False)
ap.optionDataDefaults.set("color", True)
ap.optionDataDefaults.set("stdin", False)
ap.optionDataDefaults.set("extractionPath", None)

ap.createOption('h', 'help', "Display this help text.").onOption = \
	lambda argOption, argOptionArguments, parsedArgs: \
		parsedArgs.optionData.set("help", True)
ap.createOption(None, 'no-color', "Dont' use colors in output.").onOption = \
	lambda argOption, argOptionArguments, parsedArgs: \
		parsedArgs.optionData.set("colors", False)
ap.createOption(None, 'stdin', "Read JSON data from stdin.").onOption = \
	lambda argOption, argOptionArguments, parsedArgs: \
		parsedArgs.optionData.set("stdin", True)
ap.createOption(None, 'extract', "Extract a specific part of the data object.").expectString("spath", 1).onOption = \
	lambda argOption, argOptionArguments, parsedArgs: \
		parsedArgs.optionData.set("extractionPath", jk_flexdata.FlexDataSelector(argOptionArguments[0]))

ap.createAuthor("Jürgen Knauth", "jk@binary-overflow.de")
ap.setLicense("Apache", YEAR = 2020, COPYRIGHTHOLDER = "Jürgen Knauth")

ap.createReturnCode(0, "Everything is okay.")
ap.createReturnCode(1, "An error occurred.")

#ap.createCommand("filePath", "The path of the JSON file to load") 





parsedArgs = ap.parse()
#parsedArgs.dump()

if parsedArgs.optionData["help"] or ((not parsedArgs.programArgs) and not parsedArgs.optionData["stdin"]):
	ap.showHelp()
	sys.exit(0)

"""
cmdName, cmdArgs = parsedArgs.parseNextCommand()
if cmdName is None:
	ap.showHelp()
	sys.exit(0)
"""

if parsedArgs.optionData["color"]:
	log = jk_logging.ConsoleLogger.create(logMsgFormatter = jk_logging.COLOR_LOG_MESSAGE_FORMATTER, printToStdErr = True)
else:
	log = jk_logging.ConsoleLogger.create(printToStdErr = True)

bSuccess = True

if parsedArgs.optionData["stdin"]:
	line = sys.stdin.readline()
	jsonData = json.loads(line)
else:
	for filePath in parsedArgs.programArgs:
		jsonData = jk_json.loadFromFile(filePath)

extractionPath = parsedArgs.optionData.get("extractionPath")
if extractionPath:
	assert isinstance(extractionPath, jk_flexdata.FlexDataSelector)
	_, result = extractionPath.getOne(jk_flexdata.FlexObject(jsonData))
	jsonData = result
	if isinstance(jsonData, jk_flexdata.FlexObject):
		jsonData = jsonData._toDict()

if isinstance(jsonData, str):
	retLine = "-" * 160
	ret2 = jsonData.encode("utf-8").decode('unicode_escape')
	print("\n".join([ retLine, ret2, retLine ]))
else:
	jk_json.prettyPrint(jsonData)

sys.exit(0 if bSuccess else 1)












import json
import inspect



class ObjectEncoder(json.JSONEncoder):

	def default(self, obj):
		if hasattr(obj, "to_json"):
			return self.default(obj.to_json())
		elif hasattr(obj, "__dict__"):
			d = dict(
				(key, value)
				for key, value in inspect.getmembers(obj)
				if not key.startswith("__")
				and not inspect.isabstract(value)
				and not inspect.isbuiltin(value)
				and not inspect.isfunction(value)
				and not inspect.isgenerator(value)
				and not inspect.isgeneratorfunction(value)
				and not inspect.ismethod(value)
				and not inspect.ismethoddescriptor(value)
				and not inspect.isroutine(value)
			)
			return self.default(d)
		return obj
	#

#





#!/usr/bin/env python3
# -*- coding: utf-8 -*-





from .Token import *






class SourceCodeLocation(object):

	def __init__(self, sourceID, lineNo, charPos, endLineNo, endCharPos):
		self.sourceID = sourceID
		self.lineNo = lineNo
		self.charPos = charPos
		self.endLineNo = endLineNo
		self.endCharPos = endCharPos
	#

	def __str__(self):
		if self.sourceID != None:
			return self.sourceID + "(" + str(self.lineNo + 1) + ":" + str(self.charPos + 1) + ")"
		else:
			return "(" + str(self.lineNo + 1) + ":" + str(self.charPos + 1) + ")"
	#

	def __repr__(self):
		if self.sourceID != None:
			return self.sourceID + "(" + str(self.lineNo + 1) + ":" + str(self.charPos + 1) + ")"
		else:
			return "(" + str(self.lineNo + 1) + ":" + str(self.charPos + 1) + ")"
	#

	@staticmethod
	def fromTokens(tokenA, tokenB):
		assert isinstance(tokenA, Token)
		assert isinstance(tokenB, Token)

		return SourceCodeLocation(tokenA.sourceID, tokenA.lineNo, tokenA.charPos, tokenB.endLineNo, tokenB.endCharPos)
	#

	@staticmethod
	def fromToken(token):
		assert isinstance(token, Token)

		return SourceCodeLocation(token.sourceID, token.lineNo, token.charPos, token.endLineNo, token.endCharPos)
	#

	def spanTo(self, sourceCodeLocation):
		assert isinstance(sourceCodeLocation, SourceCodeLocation)

		return SourceCodeLocation(self.sourceID, self.lineNo, self.charPos, sourceCodeLocation.endLineNo, sourceCodeLocation.endCharPos)
	#

#










#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from .TokenStream import *
from .ParserErrorException import *





class ParsingContext(object):

	def __init__(self, debuggingEnabled = True):
		assert isinstance(debuggingEnabled, bool)

		self.bDebugging = debuggingEnabled
		self.indent = -1
		self.__prefix = " " * 100
	#

	def _debugBegin(self, functionName, token):
		if self.indent > len(self.__prefix*2):
			self.__prefix += "  "
		print(self.__prefix[:self.indent*2] + functionName + " ... " + str(token))
	#

	def _debugEnd(self, functionName, bSuccess):
		if self.indent > len(self.__prefix*2):
			self.__prefix += "  "
		if bSuccess:
			print(self.__prefix[:self.indent*2] + functionName + " succeeded.")
		else:
			print(self.__prefix[:self.indent*2] + functionName + " failed.")
	#

#



def debugDecorator(func):

	def func_wrapper(self, ctx, ts):
		if ctx.bDebugging:
			ctx.indent += 1
			ctx._debugBegin(func.__name__, ts.peek())

		x = func(self, ctx, ts)

		if ctx.bDebugging:
			ctx._debugEnd(func.__name__, x[0])
			ctx.indent -= 1

		return x
	#

	return func_wrapper
#



class ParserBase(object):

	#
	# Parse the tokens. Construct an abstract syntax tree from the tokens provided by the tokenizer.
	#
	# @param	Tokenizer tokenizer		The tokenizer that provides the tokens.
	#
	def parse(self, tokenizer, bDebugging = False):
		raise Exception("Overwrite this method")
	#

#








__version__ = "0.2021.1.18"



from .schema_validator import Validator
from .SchemaParser import SchemaParser
from .SimpleSchemaGenerator import createObjectSchemaGenerator, createSchemaGenerator, \
	AbstractGenerator, \
	ObjectGenerator, ListGenerator,	StringGenerator, FloatGenerator, IntegerGenerator, BooleanGenerator



def parse(jsonDef, log = None):
	return SchemaParser.parse(jsonDef, log)
#

def loads(textToParse, log = None):
	import jk_json
	jsonDef = jk_json.loads(textToParse)
	return SchemaParser.parse(jsonDef, log)
#

def loadFromFile(filePath, log = None):
	with open(filePath, "r") as f:
		textToParse = f.read()
	import jk_json
	jsonDef = jk_json.loads(textToParse)
	return SchemaParser.parse(jsonDef, log)
#



#
# Deserialize a JSON string: Reconstruct a python data structure reading data from the specified JSON file.
#
# @param	str filePath		The path of the file to load.
# @param	bool bStrict		If ```True``` this parser sticks strictly to the JSON standard. If ```False``` C-style comments
#								are allowed and strings can be specified with single quotes and double quotes.
#								Furthermore NaN, positive and negative infinitiy is supported.
#
def loadJSONFromFileAndValidate(filePath, bStrict = False, bDebugging = False, validation = None):
	import jk_json
	from .schema_validator import Validator
	if validation:
		with open(filePath, "r", encoding="utf-8") as f:
			jsonData = jk_json.loads(f.read(), bStrict = bStrict, bDebugging = bDebugging)

		if isinstance(validation, dict):
			validator = SchemaParser.parse(validation, None)
		elif isinstance(validation, Validator):
			validator = validation
		else:
			raise Exception("Unexpected type for argument 'validation': " + repr(validation))

		if validator.validate(jsonData):
			return jsonData
		else:
			raise Exception("File " + repr(filePath) + " does not match the expected schema!")

	else:
		with open(filePath, "r", encoding="utf-8") as f:
			return jk_json.loads(f.read(), bStrict = bStrict, bDebugging = bDebugging)
#











#!/usr/bin/env python3
# -*- coding: utf-8 -*-



from .Token import *
from .TokenizerBase import *
from .SourceCodeLocation import *






class TokenStreamMark(object):

	def __init__(self, tokenStream):
		self.__pos = tokenStream.position
		self.__tokenStream = tokenStream
	#

	def resetToMark(self):
		self.__tokenStream._setPosition(self.__pos)
	#

#





class TokenStream(object):

	def __init__(self, tokens):
		self.__tokens = list(tokens)
		assert len(self.__tokens) > 0
		self.__pos = 0
		self.__maxpos = len(self.__tokens) - 1
		self.__eosToken = self.__tokens[self.__maxpos]
		assert self.__eosToken.type == "eos"
	#

	def reset(self):
		self.__pos = 0
	#

	@property
	def isEOS(self):
		return self.__pos == self.__maxpos
	#

	@property
	def position(self):
		return self.__pos
	#

	def mark(self):
		return TokenStreamMark(self)
	#

	def multiPeek(self, nCount):
		if nCount <= 0:
			raise Exception("Invalid nCount!")
		if self.__pos + nCount > self.__maxpos:
			ret = []
			for i in range(0, nCount):
				j = self.__pos + i
				if j < self.__maxpos:
					ret.append(self.__tokens[j])
				else:
					ret.append(self.__eosToken)
			return ret
		else:
			return self.__tokens[self.__pos:self.__pos + nCount]
	#

	@property
	def location(self):
		t = self.__tokens[self.__pos]
		return SourceCodeLocation.fromToken(t)

	def peek(self):
		return self.__tokens[self.__pos]
	#

	def skip(self, n = 1):
		if n < 0:
			raise Exception("Invalid n: " + str(n))
		if self.__pos + n > self.__maxpos:
			raise Exception("Skipping too far!")
		self.__pos += n
	#

	def skipAll(self, tokenType, tokenText):
		assert isinstance(tokenType, str)

		n = 0
		while True:
			t = self.__tokens[self.__pos]
			if t.type == "eos":
				return n
			if t.type != tokenType:
				return n
			if tokenText != None:
				if t.text != tokenText:
					return n
			n += 1
			self.__pos += 1
	#

	def _setPosition(self, pos):
		self.__pos = pos
	#

	def read(self):
		t = self.__tokens[self.__pos]
		if self.__pos < self.__maxpos:
			self.__pos += 1
		return t
	#

#










#!/usr/bin/python3


import json
import os
import sys
import re

import jk_exceptionhelper
import jk_console
import jk_json
import jk_flexdata
import jk_logging
from jk_typing import *
import jk_argparsing









ap = jk_argparsing.ArgsParser("jkjson", "Simple JSON processor.")

ap.optionDataDefaults.set("help", False)
ap.optionDataDefaults.set("stdin", False)
ap.optionDataDefaults.set("extractionPath", None)

ap.createOption('h', 'help', "Display this help text.").onOption = \
	lambda argOption, argOptionArguments, parsedArgs: \
		parsedArgs.optionData.set("help", True)

ap.createAuthor("Jürgen Knauth", "jk@binary-overflow.de")
ap.setLicense("Apache", YEAR = 2020, COPYRIGHTHOLDER = "Jürgen Knauth")

ap.createReturnCode(0, "Everything is okay.")
ap.createReturnCode(1, "An error occurred.")

ap.createCommand("load", "Load data from a JSON file.").expectFile("<fileName>", mustExist=True, toAbsolutePath=True) 
ap.createCommand("stdin", "Load data from STDIN.")
ap.createCommand("save", "Write data to a JSON file.").expectFile("<fileName>", toAbsolutePath=True) 
ap.createCommand("savePretty", "Write data to a JSON file.").expectFile("<fileName>", toAbsolutePath=True) 
ap.createCommand("prettyPrint", "Write the JSON data to STDOUT.")
ap.createCommand("linePrint", "Write the JSON data to STDOUT.")
ap.createCommand("print", "Canonincal print to STDOUT.")
ap.createCommand("extract", "Extract a specific part of the JSON data.").expectString("<extractionPath>", 1)





parsedArgs = ap.parse()
#parsedArgs.dump()

if parsedArgs.optionData["help"] or (len(parsedArgs.programArgs) == 0):
	ap.showHelp()
	sys.exit(0)






bSuccess = False
jsonData = None

try:

	while True:
		cmdName, cmdArgs = parsedArgs.parseNextCommand()
		if cmdName is None:
			bSuccess = True
			break

		if cmdName == "stdin":
			line = sys.stdin.readline()
			jsonData = json.loads(line)
			continue

		if cmdName == "load":
			jsonData = jk_json.loadFromFile(cmdArgs[0])
			continue

		if cmdName == "save":
			if jsonData is None:
				raise Exception("No data loaded!")
			jk_json.saveToFile(jsonData, cmdArgs[0])
			continue

		if cmdName == "savePretty":
			if jsonData is None:
				raise Exception("No data loaded!")
			jk_json.saveToFilePretty(jsonData, cmdArgs[0])
			continue

		if cmdName == "prettyPrint":
			if jsonData is None:
				raise Exception("No data loaded!")
			jk_json.prettyPrint(jsonData)
			continue

		if cmdName == "linePrint":
			if jsonData is None:
				raise Exception("No data loaded!")
			print(jk_json.dumps(jsonData))
			continue

		if cmdName == "print":
			if jsonData is None:
				raise Exception("No data loaded!")
			if isinstance(jsonData, (str,int,float,bool)):
				print(jsonData)
			else:
				jk_json.prettyPrint(jsonData)
			continue

		if cmdName == "extract":
			if jsonData is None:
				raise Exception("No data loaded!")
			extractionPath = jk_flexdata.FlexDataSelector(cmdArgs[0])
			_, result = extractionPath.getOne(jk_flexdata.FlexObject(jsonData))
			if isinstance(result, jk_flexdata.FlexObject):
				jsonData = result._toDict()
			else:
				jsonData = result
			continue

		raise Exception(repr((cmdName, cmdArgs)))

except Exception as ee:
	jk_exceptionhelper.analyseException(ee).dump()




sys.exit(0 if bSuccess else 1)












#!/usr/bin/env python3
# -*- coding: utf-8 -*-



from .SourceCodeLocation import *




class ParserErrorException(Exception):

	def __init__(self, location, message):
		assert isinstance(location, SourceCodeLocation)

		super().__init__(str(location) + " :: " + message)

		self.__location = location
		self.__message = message
	#

	@property
	def location(self):
		return self.__location
	#

	@property
	def message(self):
		return self.__message
	#

#










#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import re
from typing import List, Union

import jk_logging
from jk_testing import Assert
from jk_json.tools import *

from ._SchemaAST import _SchemaAST




# ================================================================================================================================
# ================================================================================================================================
# ================================================================================================================================



class ValidatorStackTrace(list):

	def __init__(self, errPath:str, errValue, errMsg:str):
		assert isinstance(errPath, str)
		assert isinstance(errMsg, str)
		self.append((errPath, errValue, errMsg))
	#

	def appendError(self, errPath:str, errValue, errMsg:str):
		assert isinstance(errPath, str)
		assert isinstance(errMsg, str)
		self.append((errPath, errValue, errMsg))
		return self
	#

#



# ================================================================================================================================
# ================================================================================================================================
# ================================================================================================================================



class ValidationContext(object):

	def __init__(self):
		self.allProperties = None
	#

	def derive(self):
		return ValidationContext()
	#

#



class ValidationContext2(object):

	def __init__(self, path:str = "/"):
		self.allProperties = None
		self.__path = path
	#

	@property
	def path(self):
		if self.__path == "/":
			return "/"
		else:
			return self.__path[:-1]
	#

	def deriveOnIndex(self, index:int):
		p = self.__path + str(index) + "/"
		return ValidationContext2(p)
	#

	def deriveOnPropertyName(self, propertyName:str):
		p = self.__path + propertyName + "/"
		return ValidationContext2(p)
	#

#



class AbstractElementaryValidator(object):

	def validate(self, ctx:ValidationContext, jsonData):
		raise NotImplementedError()
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		raise NotImplementedError()
	#

	def dump(self, prefix = "", writeFunction = print):
		raise NotImplementedError()
	#

#



# ================================================================================================================================
# ================================================================================================================================
# ================================================================================================================================



class _TypeSpecificValidator(AbstractElementaryValidator):

	def __init__(self, typeID):
		self.__typeID = typeID
		self._validators = []
		self._default = True
	#

	def validate(self, ctx:ValidationContext, jsonData):
		for v in self._validators:
			if not v.validate(ctx, jsonData):
				return False
		return self._default
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		for v in self._validators:
			retSuccess, retStackTrace = v.validate2(ctx, jsonData)
			if not retSuccess:
				return False, retStackTrace.appendError(ctx.path, None, "failed for type: " + ALL_TYPES_NAME_LIST[self.__typeID])

		if self._default:
			return True, None
		else:
			return False, ValidatorStackTrace(ctx.path, None, "default is: to fail (a)")
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + "TypeSpecificValidator[" + ALL_TYPES_NAME_LIST[self.__typeID] + "]:")
		prefix += "\t"
		for v in self._validators:
			v.dump(prefix, writeFunction)
		writeFunction(prefix + "default: " + str(self._default).lower())
	#

#



class _Validator(AbstractElementaryValidator):

	def __init__(self):
		self._validators = {}
		for typeID in ALL_TYPES_SET:
			self._validators[typeID] = _TypeSpecificValidator(typeID)
		self._any = []
		self._default = True
	#

	def validate(self, ctx:ValidationContext, jsonData) -> bool:
		t = getTypeIDOfValue(jsonData)
		if not self._validators[t].validate(ctx, jsonData):
			return False
		for v in self._any:
			if not v.validate(ctx, jsonData):
				return False
		return self._default
	#

	def validate2(self, ctx:ValidationContext2, jsonData) -> tuple:
		t = getTypeIDOfValue(jsonData)

		retSuccess, retStackTrace = self._validators[t].validate2(ctx, jsonData)
		if not retSuccess:
			return False, retStackTrace.appendError(ctx.path, None, "failed in category for type: " + ALL_TYPES_NAME_LIST[t])

		for v in self._any:
			retSuccess, retStackTrace = v.validate2(ctx, jsonData)
			if not retSuccess:
				return False, retStackTrace.appendError(ctx.path, None, "failed in category for all types")

		if self._default:
			return True, None
		else:
			return False, ValidatorStackTrace(ctx.path, None, "default is: to fail (b)")
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + "Validator:")
		prefix += "\t"
		for v in self._validators.values():
			v.dump(prefix, writeFunction)
		writeFunction(prefix + "any:")
		prefix2 = prefix + "\t"
		for a in self._any:
			a.dump(prefix2, writeFunction)
		writeFunction(prefix + "default: " + str(self._default).lower())
	#

#



# ================================================================================================================================
# ================================================================================================================================
# ================================================================================================================================



class Validator(object):

	def __init__(self, validator:_Validator):
		self.__validator = validator
	#

	#
	# Validate the specified JSON data.
	#
	# @return	bool bValidationResult		Returns `True` or `False`.
	#
	def validate(self, jsonData) -> bool:
		return self.__validator.validate(ValidationContext(), jsonData)
	#

	#
	# Validate the specified JSON data.
	#
	# @return	bool bValidationResult				Returns `True` or `False`.
	# @return	ValidatorStackTrace stackTrace		If validation failed returns a stack trace object. Otherwise `None` is returned.
	#
	def validate2(self, jsonData) -> tuple:
		return self.__validator.validate2(ValidationContext2(), jsonData)
	#

	#
	# Validate the specified JSON data. An exception is thrown on error.
	#
	def validateE(self, jsonData):
		bResult, errStackTrace = self.__validator.validate2(ValidationContext2(), jsonData)
		if not bResult:
			v = errStackTrace[0][1]
			if v is None:
				msg = errStackTrace[0][2]
				if msg[0].islower():
					msg = msg[0].upper() + msg[1:]
			else:
				msg = errStackTrace[0][2]
				#msg = str(v) + " " + msg
			raise Exception("JSON VALIDATION ERROR! " + msg)
	#

	def dump(self, writeFunction = print):
		self.__validator.dump("", writeFunction)
	#

#



# ================================================================================================================================
# ================================================================================================================================
# ================================================================================================================================



class _validate_any_anyOfTheseValues(AbstractElementaryValidator):

	def __init__(self, log, allowedValues:list):
		Assert.isInstance(allowedValues, list, log=log)

		self.__allowedValues = allowedValues
	#

	def validate(self, ctx:ValidationContext, jsonData):
		for v in self.__allowedValues:
			if jsonIsEqual(jsonData, v):
				return True
		return False
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		for v in self.__allowedValues:
			if jsonIsEqual(jsonData, v):
				return True, None
		return False, ValidatorStackTrace(ctx.path, jsonData, "not found in: " + repr(self.__allowedValues))
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ": " + repr(self.__allowedValues))
	#

#

class _validate_any_exactValue(AbstractElementaryValidator):

	def __init__(self, log, referenceValue):
		self.__referenceValue = referenceValue
	#

	def validate(self, ctx:ValidationContext, jsonData):
		return jsonIsEqual(jsonData, self.__referenceValue)
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		if jsonIsEqual(jsonData, self.__referenceValue):
			return True, None
		else:
			return False, ValidatorStackTrace(ctx.path, jsonData, "not equal to: " + repr(self.__referenceValue))
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ": " + repr(self.__referenceValue))
	#

#

# Check the value is a multiple of the argument value.
# int|float
class _validate_number_multipleValueOf(AbstractElementaryValidator):

	def __init__(self, log, referenceValue:Union[int,float]):
		Assert.isInstance(referenceValue, (int, float), log=log)

		self.__referenceValue = referenceValue
	#

	def validate(self, ctx:ValidationContext, jsonData):
		return (jsonData % self.__referenceValue) == 0
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		r = jsonData % self.__referenceValue
		if r == 0:
			return True, None
		else:
			return False, ValidatorStackTrace(ctx.path, jsonData, "not divisible by " + repr(self.__referenceValue) + ": "	\
				+ repr(jsonData) + " % " + repr(self.__referenceValue) + " == " + repr(r))
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ": " + repr(self.__referenceValue))
	#

#

# Check the value against an upper bound.
# int|float
class _validate_number_maximum(AbstractElementaryValidator):

	def __init__(self, log, referenceValue:Union[int,float]):
		Assert.isInstance(referenceValue, (int, float), log=log)

		self.__referenceValue = referenceValue
	#

	def validate(self, ctx:ValidationContext, jsonData):
		return jsonData <= self.__referenceValue
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		if jsonData <= self.__referenceValue:
			return True, None
		else:
			return False, ValidatorStackTrace(ctx.path, jsonData, "larger than: " + repr(self.__referenceValue))
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ": " + repr(self.__referenceValue))
	#

#

# Check the value against an lower bound.
# int|float
class _validate_number_minimum(AbstractElementaryValidator):

	def __init__(self, log, referenceValue:Union[int,float]):
		Assert.isInstance(referenceValue, (int, float), log=log)

		self.__referenceValue = referenceValue
	#

	def validate(self, ctx:ValidationContext, jsonData):
		return jsonData >= self.__referenceValue
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		if jsonData >= self.__referenceValue:
			return True, None
		else:
			return False, ValidatorStackTrace(ctx.path, jsonData, "smaller than: " + repr(self.__referenceValue))
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ": " + repr(self.__referenceValue))
	#

#

# Check the value against an upper bound.
# int|float
class _validate_number_exclusiveMaximum(AbstractElementaryValidator):

	def __init__(self, log, referenceValue:Union[int,float]):
		Assert.isInstance(referenceValue, (int, float), log=log)

		self.__referenceValue = referenceValue
	#

	def validate(self, ctx:ValidationContext, jsonData):
		return jsonData < self.__referenceValue
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		if jsonData < self.__referenceValue:
			return True, None
		else:
			return False, ValidatorStackTrace(ctx.path, jsonData, "larger than or equal to: " + repr(self.__referenceValue))
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ": " + repr(self.__referenceValue))
	#

#

# Check the value against an lower bound.
# int|float
class _validate_number_exclusiveMinimum(AbstractElementaryValidator):

	def __init__(self, log, referenceValue:Union[int,float]):
		Assert.isInstance(referenceValue, (int, float), log=log)

		self.__referenceValue = referenceValue
	#

	def validate(self, ctx:ValidationContext, jsonData):
		return jsonData > self.__referenceValue
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		if jsonData > self.__referenceValue:
			return True, None
		else:
			return False, ValidatorStackTrace(ctx.path, jsonData, "smaller than or equal to: " + repr(self.__referenceValue))
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ": " + repr(self.__referenceValue))
	#

#

# Check the size of the python value against a lower bound.
# int
class _validate_item_minLength(AbstractElementaryValidator):

	def __init__(self, log, referenceValue:int):
		Assert.isInstance(referenceValue, int, log=log)

		self.__referenceValue = referenceValue
	#

	def validate(self, ctx:ValidationContext, jsonData):
		return len(jsonData) >= self.__referenceValue
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		if len(jsonData) >= self.__referenceValue:
			return True, None
		else:
			return False, ValidatorStackTrace(ctx.path, jsonData, "length/size is " + str(len(jsonData)) + " and therefore smaller than: " + str(self.__referenceValue))
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ": " + repr(self.__referenceValue))
	#

#

# Check the size of the python value against an upper bound.
# int
class _validate_item_maxLength(AbstractElementaryValidator):

	def __init__(self, log, referenceValue:int):
		Assert.isInstance(referenceValue, int, log=log)

		self.__referenceValue = referenceValue
	#

	def validate(self, ctx:ValidationContext, jsonData):
		return len(jsonData) <= self.__referenceValue
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		if len(jsonData) <= self.__referenceValue:
			return True, None
		else:
			return False, ValidatorStackTrace(ctx.path, jsonData, "length/size is " + str(len(jsonData)) + " and therefore greater than: " + str(self.__referenceValue))
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ": " + repr(self.__referenceValue))
	#

#

class _validate_item_mustBeNoneOrEmpty(AbstractElementaryValidator):

	def __init__(self, log):
		pass
	#

	def validate(self, ctx:ValidationContext, jsonData):
		return (jsonData is None) or (len(jsonData) == 0)
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		if (jsonData is None) or (len(jsonData) == 0):
			return True, None
		else:
			return False, ValidatorStackTrace(ctx.path, jsonData, "there is data")
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__)
	#

#

class _validate_item_mustNotBeNoneNorEmpty(AbstractElementaryValidator):

	def __init__(self, log):
		pass
	#

	def validate(self, ctx:ValidationContext, jsonData):
		return (jsonData != None) and (len(jsonData) > 0)
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		if (jsonData != None) and (len(jsonData) > 0):
			return True, None
		else:
			return False, ValidatorStackTrace(ctx.path, jsonData, "there is no data")
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__)
	#

#

class _validate_item_mustNotBeNone(AbstractElementaryValidator):

	def __init__(self, log):
		pass
	#

	def validate(self, ctx:ValidationContext, jsonData):
		return jsonData != None
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		if jsonData != None:
			return True, None
		else:
			return False, ValidatorStackTrace(ctx.path, jsonData, "there is no data")
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__)
	#

#

# Check the string against the argument pattern.
# string_pattern
class _validate_string_pattern(AbstractElementaryValidator):

	def __init__(self, log, referencePatternStr:str):
		Assert.isInstance(referencePatternStr, str, log=log)

		self.__referencePatternStr = referencePatternStr
		self.__referencePattern = re.compile(referencePatternStr)
	#

	def validate(self, ctx:ValidationContext, jsonData):
		return self.__referencePattern.search(jsonData) != None
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		if self.__referencePattern.search(jsonData) != None:
			return True, None
		else:
			return False, ValidatorStackTrace(ctx.path, jsonData, "does not match pattern: " + repr(self.__referencePatternStr))
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ": " + repr(self.__referencePatternStr))
	#

#

# Check if all array elements match against the specified validator.
# AbstractElementaryValidator
class _validate_array_checkIfAllItemsMatchValidator(AbstractElementaryValidator):

	def __init__(self, log, validator:AbstractElementaryValidator):
		Assert.isInstance(validator, AbstractElementaryValidator, log=log)

		self.__validator = validator
	#

	def validate(self, ctx:ValidationContext, jsonData):
		for item in jsonData:
			ctx2 = ctx.derive()
			if not self.__validator.validate(ctx2, item):
				return False
		return True
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		i = 0
		for item in jsonData:
			ctx2 = ctx.deriveOnIndex(i)
			retSuccess, retStackTrace = self.__validator.validate2(ctx2, item)
			if not retSuccess:
				return False, retStackTrace.appendError(ctx.path, None, "an item does not validate")
			i += 1
		return True, None
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ":")
		writeFunction(prefix + "\tvalidator:")
		self.__validator.dump(prefix + "\t\t", writeFunction)
	#

#

# Check if at least one array elements matches against the specified validator.
# AbstractElementaryValidator
class _validate_array_checkIfAnyItemMatchesValidator(AbstractElementaryValidator):

	def __init__(self, log, validator:AbstractElementaryValidator):
		Assert.isInstance(validator, AbstractElementaryValidator, log=log)

		self.__validator = validator
	#

	def validate(self, ctx:ValidationContext, jsonData):
		for item in jsonData:
			ctx2 = ctx.derive()
			if self.__validator.validate(ctx2, item):
				return True
		return False
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		i = 0
		for item in jsonData:
			ctx2 = ctx.deriveOnIndex(i)
			retSuccess, retStackTrace = self.__validator.validate2(ctx2, item)
			if retSuccess:
				return True, None
			i += 1
		return False, retStackTrace.appendError(ctx.path, None, "no item validates")
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ":")
		writeFunction(prefix + "\tvalidator:")
		self.__validator.dump(prefix + "\t\t", writeFunction)
	#

#

# Check if all array elements match against the specified list of validators.
# AbstractElementaryValidator[]
class _validate_array_checkAllItemsMatchListOfValidators(AbstractElementaryValidator):

	def __init__(self, log, validators:List[AbstractElementaryValidator]):
		Assert.isInstance(validators, list, log=log)
		for validator in validators:
			Assert.isInstance(validator, AbstractElementaryValidator, log=log)

		self.__validators = validators
		self.__count = len(validators)
	#

	def validate(self, ctx:ValidationContext, jsonData):
		#if len(jsonData) != self.__count:
		#	return False

		for jsonItem, validator in zip(jsonData, self.__validators):
			ctx2 = ctx.derive()
			if not validator.validate(ctx2, jsonItem):
				return False
		return True
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		#if len(jsonData) != self.__count:
		#	return False, ValidatorStackTrace(ctx.path, jsonData, str(self.__count) + " items expected, " + str(len(jsonData)) + " found")

		i = 0
		for jsonItem, validator in zip(jsonData, self.__validators):
			ctx2 = ctx.deriveOnIndex(i)
			retSuccess, retStackTrace = validator.validate2(ctx2, jsonItem)
			if not retSuccess:
				return False, retStackTrace.appendError(ctx.path, None, "validator at index " + str(i) + " failed.")
			i += 1
		return True, None
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ":")
		writeFunction(prefix + "\tvalidators:")
		for v in self.__validators:
			v.dump(prefix + "\t\t", writeFunction)
	#

#

# Check if all array elements match against the specified list of validators.
# AbstractElementaryValidator[], AbstractElementaryValidator
class _validate_array_checkAllItemsMatchListOfValidatorsWithExtraValidator(AbstractElementaryValidator):

	def __init__(self, log, validators:List[AbstractElementaryValidator], extraValidator:AbstractElementaryValidator):
		Assert.isInstance(validators, list, log=log)
		for validator in validators:
			Assert.isInstance(validator, AbstractElementaryValidator, log=log)
		Assert.isInstance(extraValidator, AbstractElementaryValidator, log=log)

		self.__validators = validators
		self.__extraValidator = extraValidator
		self.__count = len(validators)
	#

	def validate(self, ctx:ValidationContext, jsonData):
		if len(jsonData) < self.__count:
			return False
		extraJsonData = jsonData[self.__count:]
		jsonData = jsonData[0:self.__count]

		for jsonItem, validator in zip(jsonData, self.__validators):
			ctx2 = ctx.derive()
			if not validator.validate(ctx2, jsonItem):
				return False

		for jsonItem in extraJsonData:
			ctx2 = ctx.derive()
			if not self.__extraValidator.validate(ctx2, jsonItem):
				return False

		return True
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		#if len(jsonData) < self.__count:
		#	return False, ValidatorStackTrace(ctx.path, jsonData, "at least " + str(self.__count) + " items expected, " + str(len(jsonData)) + " found")

		extraJsonData = jsonData[self.__count:]
		jsonData = jsonData[0:self.__count]

		i = 0
		for jsonItem, validator in zip(jsonData, self.__validators):
			ctx2 = ctx.deriveOnIndex(i)
			retSuccess, retStackTrace = validator.validate2(ctx2, jsonItem)
			if not retSuccess:
				return False, retStackTrace.appendError(ctx.path, None, "validator failed at index " + str(i))
			i += 1

		for jsonItem in extraJsonData:
			ctx2 = ctx.deriveOnIndex(i)
			retSuccess, retStackTrace = self.__extraValidator.validate2(ctx2, jsonItem)
			if not retSuccess:
				return False, retStackTrace.appendError(ctx.path, None, "extra validator failed at index " + str(i))
			i += 1

		return True, None
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ":")
		writeFunction(prefix + "\tvalidators:")
		for v in self.__validators:
			v.dump(prefix + "\t\t", writeFunction)
		writeFunction(prefix + "\textraValidator:")
		self.__extraValidator.dump(prefix + "\t\t", writeFunction)
	#

#

# Check if all array elements match against the specified list of validators.
# AbstractElementaryValidator[], AbstractElementaryValidator[]
class _validate_array_checkAllItemsMatchListOfValidatorsWithExtraListOfValidators(AbstractElementaryValidator):

	def __init__(self, log, validators:List[AbstractElementaryValidator], extraValidators:List[AbstractElementaryValidator]):
		Assert.isInstance(validators, list, log=log)
		for validator in validators:
			Assert.isInstance(validator, AbstractElementaryValidator, log=log)
		Assert.isInstance(extraValidators, list, log=log)
		for validator in extraValidators:
			Assert.isInstance(validator, AbstractElementaryValidator, log=log)

		self.__validators = validators
		self.__extraValidators = extraValidators
		self.__count = len(validators)
		self.__extraCount = len(self.__extraValidators)
	#

	def validate(self, ctx:ValidationContext, jsonData):
		if len(jsonData) < self.__count:
			return False
		extraJsonData = jsonData[self.__count:]
		jsonData = jsonData[0:self.__count]

		if len(extraJsonData) < self.__extraCount:
			return False

		for jsonItem, validator in zip(jsonData, self.__validators):
			ctx2 = ctx.derive()
			if not validator.validate(ctx2, jsonItem):
				return False

		for jsonItem, validator in zip(extraJsonData, self.__extraValidators):
			ctx2 = ctx.derive()
			if not validator.validate(ctx2, jsonItem):
				return False

		return True
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		#if len(jsonData) < self.__count:
		#	return False, ValidatorStackTrace(ctx.path, jsonData, "at least " + str(self.__count) + " items expected, " + str(len(jsonData)) + " found")

		extraJsonData = jsonData[self.__count:]
		jsonData = jsonData[0:self.__count]

		i = 0
		for jsonItem, validator in zip(jsonData, self.__validators):
			ctx2 = ctx.deriveOnIndex(i)
			retSuccess, retStackTrace = validator.validate2(ctx2, jsonItem)
			if not retSuccess:
				return False, retStackTrace.appendError(ctx.path, None, "validator failed at index " + str(i))
			i += 1

		for jsonItem, validator in zip(extraJsonData, self.__extraValidators):
			ctx2 = ctx.deriveOnIndex(i)
			retSuccess, retStackTrace = validator.validate2(ctx2, jsonItem)
			if not retSuccess:
				return False, retStackTrace.appendError(ctx.path, None, "extra validator failed at index " + str(i))
			i += 1

		return True, None
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ":")
		writeFunction(prefix + "\tvalidators:")
		for v in self.__validators:
			v.dump(prefix + "\t\t", writeFunction)
		writeFunction(prefix + "\textraValidators:")
		for v in self.__extraValidators:
			v.dump(prefix + "\t\t", writeFunction)
	#

#

# Check if no array element is equal to another array element.
class _validate_array_itemsMustBeUnique(AbstractElementaryValidator):

	def __init__(self, log):
		pass
	#

	def validate(self, ctx:ValidationContext, jsonData):
		n = len(jsonData)
		if n < 2:
			return True
		for i in range(0, n - 1):
			v1 = jsonData[i]
			t1 = getTypeIDOfValue(v1)
			for j in range(i + 1, n):
				v2 = jsonData[j]
				t2 = getTypeIDOfValue(v2)
				if t1 != t2:
					continue
				if jsonIsEqual(v1, v2):
					return False
		return True
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		n = len(jsonData)
		if n < 2:
			return True, None
		for i in range(0, n - 1):
			v1 = jsonData[i]
			t1 = getTypeIDOfValue(v1)
			for j in range(i + 1, n):
				v2 = jsonData[j]
				t2 = getTypeIDOfValue(v2)
				if t1 != t2:
					continue
				if jsonIsEqual(v1, v2):
					return False, ValidatorStackTrace(ctx.path, jsonData, "item at " + int(i) + " and " + int(j) + " are equal")
		return True, None
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__)
	#

#

# Check if at least one array element matches the reference.
# (any)
class _validate_array_mustContainExactValue(AbstractElementaryValidator):

	def __init__(self, log, referenceValue):
		self.__referenceValue = referenceValue
	#

	def validate(self, ctx:ValidationContext, jsonData):
		for v in jsonData:
			if jsonIsEqual(v, self.__referenceValue):
				return True
		return False
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		for v in jsonData:
			if jsonIsEqual(v, self.__referenceValue):
				return True, None
		return False, ValidatorStackTrace(ctx.path, jsonData, "no such item: " + repr(self.__referenceValue))
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ": " + repr(self.__referenceValue))
	#

#

# AbstractElementaryValidator
class _validate_not(AbstractElementaryValidator):

	def __init__(self, log, validator:AbstractElementaryValidator):
		Assert.isInstance(validator, AbstractElementaryValidator, log=log)

		self.__validator = validator
	#

	def validate(self, ctx:ValidationContext, jsonData):
		return not self.__validator.validate(ctx, jsonData)
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		if not self.__validator.validate(ctx, jsonData):
			return True, None
		else:
			return False, ValidatorStackTrace(ctx.path, jsonData, "validator validated successfully")
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ":")
		writeFunction(prefix + "\tvalidator:")
		self.__validator.dump(prefix + "\t\t", writeFunction)
	#

#

# AbstractElementaryValidator, AbstractElementaryValidator, AbstractElementaryValidator
class _validate_if(AbstractElementaryValidator):

	def __init__(self, log, validatorIf:AbstractElementaryValidator, validatorThen:AbstractElementaryValidator, validatorElse:AbstractElementaryValidator):
		Assert.isInstance(validatorIf, AbstractElementaryValidator, log=log)
		if validatorThen:
			Assert.isInstance(validatorThen, AbstractElementaryValidator, log=log)
		if validatorElse:
			Assert.isInstance(validatorElse, AbstractElementaryValidator, log=log)

		self.__validatorIf = validatorIf
		self.__validatorThen = validatorThen
		self.__validatorElse = validatorElse
	#

	def validate(self, ctx:ValidationContext, jsonData):
		if self.__validatorIf.validate(ctx, jsonData):
			if self.__validatorThen:
				return self.__validatorThen.validate(ctx, jsonData)
			else:
				return True
		else:
			if self.__validatorElse:
				return self.__validatorElse.validate(ctx, jsonData)
			else:
				return True
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		if self.__validatorIf.validate(ctx, jsonData):
			if self.__validatorThen:
				retSuccess, retStackTrace = self.__validatorThen.validate(ctx, jsonData)
				if retSuccess:
					return True, None
				else:
					return False, retStackTrace.appendError(ctx.path, None, "then-validator failed.")
			else:
				return True
		else:
			if self.__validatorElse:
				retSuccess, retStackTrace = self.__validatorElse.validate(ctx, jsonData)
				if retSuccess:
					return True, None
				else:
					return False, retStackTrace.appendError(ctx.path, None, "else-validator failed.")
			else:
				return True
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ":")
		writeFunction(prefix + "\tvalidatorIF:")
		self.__validatorIf.dump(prefix + "\t\t", writeFunction)
		if self.__validatorThen:
			writeFunction(prefix + "\tvalidatorTHEN:")
			self.__validatorThen.dump(prefix + "\t\t", writeFunction)
		if self.__validatorElse:
			writeFunction(prefix + "\tvalidatorELSE:")
			self.__validatorElse.dump(prefix + "\t\t", writeFunction)
	#

#

# Check if all validators evaluate to true
# AbstractElementaryValidator[]
class _validate_allOf(AbstractElementaryValidator):

	def __init__(self, log, validators:List[AbstractElementaryValidator]):
		Assert.isInstance(validators, list, log=log)
		for validator in validators:
			Assert.isInstance(validator, AbstractElementaryValidator, log=log)

		self.__validators = validators
	#

	def validate(self, ctx:ValidationContext, jsonData):
		for v in self.__validators:
			if not v.validate(ctx, jsonData):
				return False
		return True
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		for v in self.__validators:
			retSuccess, retStackTrace = v.validate2(ctx, jsonData)
			if not retSuccess:
				return False, retStackTrace.appendError(ctx.path, None, "validator failed")
		return True, None
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ":")
		writeFunction(prefix + "\tvalidators:")
		for v in self.__validators:
			v.dump(prefix + "\t\t", writeFunction)
	#

#

# Check if at least one validator evaluates to true
# AbstractElementValidator[]
class _validate_anyOf(AbstractElementaryValidator):

	def __init__(self, log, validators:List[AbstractElementaryValidator]):
		Assert.isInstance(validators, list, log=log)
		for validator in validators:
			Assert.isInstance(validator, AbstractElementaryValidator, log=log)

		self.__validators = validators
	#

	def validate(self, ctx:ValidationContext, jsonData):
		for v in self.__validators:
			if v.validate(ctx, jsonData):
				return True
		return False
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		for v in self.__validators:
			retSuccess, retStackTrace = v.validate2(ctx, jsonData)
			if retSuccess:
				return True, None
		return False, retStackTrace.appendError(ctx.path, None, "no validator succeeded")
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ":")
		writeFunction(prefix + "\tvalidators:")
		for v in self.__validators:
			v.dump(prefix + "\t\t", writeFunction)
	#

#

# Check if exactly one validator evaluates to true
# AbstractElementValidator[]
class _validate_oneOf(AbstractElementaryValidator):

	def __init__(self, log, validators:List[AbstractElementaryValidator]):
		Assert.isInstance(validators, list, log=log)
		for validator in validators:
			Assert.isInstance(validator, AbstractElementaryValidator, log=log)

		self.__validators = validators
	#

	def validate(self, ctx:ValidationContext, jsonData):
		n = 0
		for v in self.__validators:
			if v.validate(ctx, jsonData):
				if n == 0:
					n += 1
				else:
					return False
		return n == 1
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		n = 0
		for v in self.__validators:
			if v.validate(ctx, jsonData):
				if n == 0:
					n += 1
				else:
					return False, ValidatorStackTrace(ctx.path, jsonData, "another validator already succeeded")
		if n == 1:
			return True, None
		else:
			return False, ValidatorStackTrace(ctx.path, jsonData, "no validator succeeded")
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ":")
		writeFunction(prefix + "\tvalidators:")
		for v in self.__validators:
			v.dump(prefix + "\t\t", writeFunction)
	#

#

# Check if all property names match against the specified validator.
# AbstractElementaryValidator
class _validate_object_allPropertyNamesMatchValidator(AbstractElementaryValidator):

	def __init__(self, log, validator:AbstractElementaryValidator):
		Assert.isInstance(validator, AbstractElementaryValidator, log=log)

		self.__validator = validator
	#

	def validate(self, ctx:ValidationContext, jsonData):
		for propertyName in jsonData.keys():
			if not self.__validator.validate(ctx, propertyName):
				return False
		return True
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		for propertyName in jsonData.keys():
			retSuccess, retStackTrace = self.__validator.validate2(ctx, propertyName)
			if not retSuccess:
				return False, retStackTrace.appendError(ctx.path, propertyName, "property name failed to validate")
		return True, None
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ":")
		writeFunction(prefix + "\tvalidator:")
		self.__validator.dump(prefix + "\t\t", writeFunction)
	#

#

class _validate_object_otherPropertyMustExist(AbstractElementaryValidator):

	def __init__(self, log, propertyNameToExpectedPropertyNamesMap:dict):
		Assert.isInstance(propertyNameToExpectedPropertyNamesMap, dict, log=log)
		for key, value in propertyNameToExpectedPropertyNamesMap.items():
			Assert.isInstance(key, str, log=log)
			Assert.isInstance(value, list, log=log)
			for item in value:
				Assert.isInstance(item, str, log=log)

		self.__propertyNameToExpectedPropertyNamesMap = propertyNameToExpectedPropertyNamesMap
	#

	def validate(self, ctx:ValidationContext, jsonData):
		for propertyName, expectedOtherProperties in self.__propertyNameToExpectedPropertyNamesMap.items():
			if propertyName in jsonData:
				for p in expectedOtherProperties:
					if not p in jsonData:
						return False
		return True
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		for propertyName, expectedOtherProperties in self.__propertyNameToExpectedPropertyNamesMap.items():
			if propertyName in jsonData:
				for p in expectedOtherProperties:
					if not p in jsonData:
						return False, ValidatorStackTrace(ctx.path, jsonData, repr(propertyName) + " exists, so expecting " + repr(expectedOtherProperties)
							+ " from where " + repr(p) + " did not exist")
		return True, None
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ":")
		writeFunction(prefix + "\tpropertyNameToExpectedPropertyNamesMap:")
		for propertyName, expectedOtherProperties in self.__propertyNameToExpectedPropertyNamesMap.items():
			writeFunction(prefix + "\t\t" + repr(propertyName) + ": " + repr(expectedOtherProperties))
	#

#

class _validate_object_objectMustMatchSchema(AbstractElementaryValidator):

	def __init__(self, log, propertyNameToSchemaMap:dict):
		Assert.isInstance(propertyNameToSchemaMap, dict, log=log)
		for key, value in propertyNameToSchemaMap.items():
			Assert.isInstance(key, str, log=log)
			Assert.isInstance(value, AbstractElementaryValidator, log=log)

		self.__propertyNameToSchemaMap = propertyNameToSchemaMap
	#

	def validate(self, ctx:ValidationContext, jsonData):
		for propertyName, validator in self.__propertyNameToSchemaMap.items():
			if propertyName in jsonData:
				return validator.validate(ctx, jsonData)
		return True
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		for propertyName, validator in self.__propertyNameToSchemaMap.items():
			if propertyName in jsonData:
				retSuccess, retStackTrace = validator.validate2(ctx, jsonData)
				if not retSuccess:
					return False, retStackTrace.appendError(ctx.path, None, repr(propertyName) + " exists, so expecting property related schema to match object but that failed")
		return True, None
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ":")
		writeFunction(prefix + "\tpropertyNameToSchemaMap:")
		for propertyName, validator in self.__propertyNameToSchemaMap.items():
			writeFunction(prefix + "\t\t" + repr(propertyName) + ":")
			validator.dump(prefix + "\t\t\t", writeFunction)
	#

#

# dict<str,_SchemaAST>
# modifies: ctx.allProperties
class _validate_object_propertiesMustMatchSchema(AbstractElementaryValidator):

	def __init__(self, log, propertyNameToSchemaMap:dict):
		Assert.isInstance(propertyNameToSchemaMap, dict, log=log)
		for key, value in propertyNameToSchemaMap.items():
			Assert.isInstance(key, str, log=log)
			Assert.isInstance(value, AbstractElementaryValidator, log=log)

		self.__propertyNameToSchemaMap = propertyNameToSchemaMap
	#

	def validate(self, ctx:ValidationContext, jsonData):
		for propertyName, validator in self.__propertyNameToSchemaMap.items():
			ctx2 = ctx.derive()
			if propertyName in jsonData:
				ctx.allProperties.remove(propertyName)
				v = jsonData[propertyName]
				if not validator.validate(ctx2, v):
					return False
			else:
				return True
		return True
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		for propertyName, validator in self.__propertyNameToSchemaMap.items():
			if propertyName in jsonData:
				ctx.allProperties.remove(propertyName)
				ctx2 = ctx.deriveOnPropertyName(propertyName)
				v = jsonData[propertyName]
				retSuccess, retStackTrace = validator.validate2(ctx2, v)
				if not retSuccess:
					return False, retStackTrace.appendError(ctx.path, None, repr(propertyName) + " exists, so expecting property related schema to match property value but that failed")
		return True, None
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ":")
		writeFunction(prefix + "\tpropertyNameToSchemaMap:")
		for propertyName, validator in self.__propertyNameToSchemaMap.items():
			writeFunction(prefix + "\t\t" + repr(propertyName) + ":")
			validator.dump(prefix + "\t\t\t", writeFunction)
	#

#

# list<(regexstr,_SchemaAST)>
# modifies: ctx.allProperties
class _validate_object_regexPropertiesMustMatchSchema(AbstractElementaryValidator):

	def __init__(self, log, propertyNamePatternAndValidatorList:list):
		Assert.isInstance(propertyNamePatternAndValidatorList, list, log=log)
		for propertyNamePattern, validator in propertyNamePatternAndValidatorList:
			Assert.isInstance(propertyNamePattern, str, log=log)
			Assert.isInstance(validator, AbstractElementaryValidator, log=log)

		self.__recordList =	\
			[ (p, re.compile(p), v) for (p, v) in propertyNamePatternAndValidatorList ]
	#

	def validate(self, ctx:ValidationContext, jsonData):
		for propertyName in jsonData:
			for propertyNamePattern, rePattern, validator in self.__recordList:
				if rePattern.search(propertyName):
					ctx.allProperties.remove(propertyName)
					ctx2 = ctx.derive()
					v = jsonData[propertyName]
					retSuccess = validator.validate(ctx2, v)
					if not retSuccess:
						return False
		return True
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		for propertyName in jsonData:
			for propertyNamePattern, rePattern, validator in self.__recordList:
				if rePattern.search(propertyName):
					ctx.allProperties.remove(propertyName)
					ctx2 = ctx.deriveOnPropertyName(propertyName)
					v = jsonData[propertyName]
					retSuccess, retStackTrace = validator.validate2(ctx2, v)
					if not retSuccess:
						return False, retStackTrace.appendError(ctx.path, None, repr(propertyName) + " matches " + propertyNamePattern + " but validation of property value failed")
		return True, None
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ":")
		writeFunction(prefix + "\trecordList:")
		for propertyNamePattern, rePattern, validator in self.__recordList:
			writeFunction(prefix + "\t\t" + repr(propertyNamePattern) + ":")
			validator.dump(prefix + "\t\t\t", writeFunction)
	#

#

# Check that certain properties exist
# str[]
class _validate_object_mustHaveProperties(AbstractElementaryValidator):

	def __init__(self, log, propertyNames:list):
		Assert.isInstance(propertyNames, list, log=log)
		for key in propertyNames:
			Assert.isInstance(key, str, log=log)

		self.__propertyNames = propertyNames
	#

	def validate(self, ctx:ValidationContext, jsonData):
		for propertyName in self.__propertyNames:
			if propertyName not in jsonData:
				return False
		return True
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		for propertyName in self.__propertyNames:
			if propertyName not in jsonData:
				return False, ValidatorStackTrace(ctx.path, jsonData, repr(propertyName) + " does not exist")
		return True, None
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ":")
		writeFunction(prefix + "\tpropertyNames: " + repr(self.__propertyNames))
	#

#

# modifies: ctx.allProperties
class _validate_object_collectAllPropertyNames(AbstractElementaryValidator):

	def __init__(self, log):
		pass
	#

	def validate(self, ctx:ValidationContext, jsonData):
		ctx.allProperties = set(jsonData.keys())
		return True
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		ctx.allProperties = set(jsonData.keys())
		return True, None
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ":")
	#

#

# Check if all additional properties match against the specified validator.
# AbstractElementaryValidator
class _validate_object_additionalPropertiesMatchValidator(AbstractElementaryValidator):

	def __init__(self, log, validator:AbstractElementaryValidator):
		Assert.isInstance(validator, AbstractElementaryValidator, log=log)

		self.__validator = validator
	#

	def validate(self, ctx:ValidationContext, jsonData):
		if ctx.allProperties is None:
			raise Exception()

		for propertyName in ctx.allProperties:
			ctx2 = ctx.derive()
			if not self.__validator.validate(ctx2, jsonData[propertyName]):
				return False
		return True
	#

	def validate2(self, ctx:ValidationContext2, jsonData):
		if ctx.allProperties is None:
			raise Exception()

		for propertyName in ctx.allProperties:
			ctx2 = ctx.deriveOnPropertyName(propertyName)
			retSuccess, retStackTrace = self.__validator.validate2(ctx2, jsonData[propertyName])
			if not retSuccess:
				return False, retStackTrace.appendError(ctx.path, jsonData, "property failed to validate")
		return True, None
	#

	def dump(self, prefix = "", writeFunction = print):
		writeFunction(prefix + self.__class__.__name__ + ":")
		writeFunction(prefix + "\tvalidator:")
		self.__validator.dump(prefix + "\t\t", writeFunction)
	#

#





# TODO: check argument types as well, not only the number of arguments

ANY_VALIDATORS = {
	("any_allowedValues", _validate_any_anyOfTheseValues, 1),
}

NUMBER_VALIDATORS = (
	("number_minimum", _validate_number_minimum, 1),
	("number_exclusiveMinimum", _validate_number_exclusiveMinimum, 1),
	("number_maximum", _validate_number_maximum, 1),
	("number_exclusiveMaximum", _validate_number_exclusiveMaximum, 1),
	("number_multipleValueOf", _validate_number_multipleValueOf, 1),
)

STRING_VALIDATORS = (
	("string_minLength", _validate_item_minLength, 1),
	("string_maxLength", _validate_item_maxLength, 1),
	("string_pattern", _validate_string_pattern, 1),
)

ARRAY_VALIDATORS = (
	("array_minItems", _validate_item_minLength, 1),
	("array_maxItems", _validate_item_maxLength, 1),
	("array_mustContainExactValue", _validate_array_mustContainExactValue, 1),
	("array_itemsMustBeUnique", _validate_array_itemsMustBeUnique, 0),
)

OPS_VALIDATORS = {
	("op_not", _validate_not, 1),
	("op_anyOf", _validate_anyOf, 1),
	("op_allOf", _validate_allOf, 1),
	("op_oneOf", _validate_oneOf, 1),
}

EXCLUSIVE_OPS = [
	"op_if", "op_not", "op_anyOf", "op_oneOf", "op_allOf"
]

OBJECT_VALIDATORS = (
	("object_requiredProperties", _validate_object_mustHaveProperties, 1),
	("object_minProperties", _validate_item_minLength, 1),
	("object_maxProperties", _validate_item_maxLength, 1),
	("object_propertyDependencyObjectMustMatchSchema", _validate_object_objectMustMatchSchema, 1),
	("object_propertyDependencyOtherPropertyMustExist", _validate_object_otherPropertyMustExist, 1),
	("object_propertyNames", _validate_object_allPropertyNamesMatchValidator, 1),
	("object_properties", _validate_object_propertiesMustMatchSchema, 1),							# must be placed before object_additionalProperties
	("object_patternProperties", _validate_object_regexPropertiesMustMatchSchema, 1),				# must be placed before object_additionalProperties
	("object_additionalProperties", _validate_object_additionalPropertiesMatchValidator, 1),		# must be placed after object_properties and after object_patternProperties
)






def compileAstList(astList:list, log:jk_logging.AbstractLogger):
	Assert.isInstance(log, jk_logging.AbstractLogger, log=log)

	if astList is None:
		return None
	Assert.isInstance(astList, list)
	return [ compileAst(x, log) for x in astList ]
#



def tryCompile(something, log:jk_logging.AbstractLogger):
	if something is None:
		return None
	elif isinstance(something, _SchemaAST):
		return compileAst(something, log)
	elif isinstance(something, (tuple, list)):
		return [ tryCompile(item, log) for item in something ]
	elif isinstance(something, dict):
		return { key: tryCompile(value, log) for (key, value) in something.items() }
	else:
		return something
#



def __compileAccordingToDefinitions(ast:_SchemaAST, typeIDorTypeIDs:Union[None,int,List[int]], listOfValidatorDefs:list, ret:_Validator, log:jk_logging.AbstractLogger):
	for (astVarName, validatorClass, nArgs) in listOfValidatorDefs:
		bAppendValidator = False

		if nArgs == 0:
			a = getattr(ast, astVarName, None)
			if a != None:
				v = validatorClass(log)
				bAppendValidator = True
		elif nArgs == 1:
			a = getattr(ast, astVarName, None)
			if a != None:
				a = tryCompile(a, log)
				v = validatorClass(log, a)
				bAppendValidator = True
		else:
			a = getattr(ast, astVarName, None)
			if a != None:
				a = tryCompile(a, log)
				v = validatorClass(log, *a)
				bAppendValidator = True

		if bAppendValidator:
			if typeIDorTypeIDs is None:
				ret._any.append(v)
			elif isinstance(typeIDorTypeIDs, (tuple, list)):
				for typeID in typeIDorTypeIDs:
					ret._validators[typeID]._validators.append(v)
			else:
				ret._validators[typeIDorTypeIDs]._validators.append(v)
#



def compileAst(ast:_SchemaAST, log:jk_logging.AbstractLogger):
	Assert.isInstance(log, jk_logging.AbstractLogger, log=log)

	if ast is None:
		return None
	Assert.isInstance(ast, _SchemaAST, log=log)

	ret = _Validator()

	# -------- boolean --------

	if ast.always != None:
		# this overrides everything
		ret._default = ast.always
		return ret

	# -------- ops --------

	lastOpVarName = None
	for astOpVarName in EXCLUSIVE_OPS:
		a = getattr(ast, astOpVarName, None)
		if a != None:
			if lastOpVarName:
				raise Exception("Can't compile " + astOpVarName + " as " + lastOpVarName + " has already been defined!")
			lastOpVarName = astOpVarName

	if ast.op_if:
		ret._any.append(_validate_if(log,
			compileAst(ast.op_if, log),
			compileAst(ast.op_then, log),
			compileAst(ast.op_else, log)))

	__compileAccordingToDefinitions(ast, None, OPS_VALIDATORS, ret, log)

	# -------- any --------

	__compileAccordingToDefinitions(ast, None, ANY_VALIDATORS, ret, log)

	if ast.any_mustBeExactValue:
		ret._any.append(_validate_any_exactValue(log, ast.any_mustBeExactValue[0]))

	# -------- number --------

	__compileAccordingToDefinitions(ast, (TYPE_INT, TYPE_FLOAT), NUMBER_VALIDATORS, ret, log)

	# -------- string --------

	__compileAccordingToDefinitions(ast, TYPE_STR, STRING_VALIDATORS, ret, log)

	# -------- array --------

	__compileAccordingToDefinitions(ast, TYPE_ARRAY, ARRAY_VALIDATORS, ret, log)

	if ast.array_itemsMustBeNone:
		ret._validators[TYPE_ARRAY]._validators.append(_validate_item_mustBeNoneOrEmpty(log))
	elif ast.array_itemsMustNotBeNone:
		ret._validators[TYPE_ARRAY]._validators.append(_validate_item_mustNotBeNone(log))
	elif ast.array_items != None:
		if isinstance(ast.array_items, _SchemaAST):
			ret._validators[TYPE_ARRAY]._validators.append(_validate_array_checkIfAllItemsMatchValidator(log,
				compileAst(ast.array_items, log)))
		else:
			# assert isinstance(ast.array_items, list)
			Assert.isInstance(ast.array_items, list, log=log)
			if ast.array_additionalItems != None:
				if isinstance(ast.array_additionalItems, bool):
					if not ast.array_additionalItems:
						ret._validators[TYPE_ARRAY]._validators.append(_validate_array_checkAllItemsMatchListOfValidatorsWithExtraValidator(log,
							compileAstList(ast.array_items, log),
							[]))
				elif isinstance(ast.array_additionalItems, _SchemaAST):
					ret._validators[TYPE_ARRAY]._validators.append(_validate_array_checkAllItemsMatchListOfValidatorsWithExtraValidator(log,
						compileAstList(ast.array_items, log),
						compileAst(ast.array_additionalItems, log)))
				else:
					# assert isinstance(ast.array_additionalItems, list)
					Assert.isInstance(ast.array_additionalItems, list, log=log)
					ret._validators[TYPE_ARRAY]._validators.append(_validate_array_checkAllItemsMatchListOfValidatorsWithExtraListOfValidators(log,
						compileAstList(ast.array_items, log),
						compileAstList(ast.array_additionalItems, log)))
			else:
				ret._validators[TYPE_ARRAY]._validators.append(_validate_array_checkAllItemsMatchListOfValidators(log,
					compileAstList(ast.array_items, log)))

	# -------- objects --------

	if ast.object_properties or ast.object_patternProperties:
		ret._validators[TYPE_OBJECT]._validators.append(_validate_object_collectAllPropertyNames(log))

	__compileAccordingToDefinitions(ast, TYPE_OBJECT, OBJECT_VALIDATORS, ret, log)

	# -------- types --------

	if ast.any_allowedTypes:
		# only let the specified types pass.
		typeIDs = set(ast.any_allowedTypes)
		extraTypeIDs = ALL_TYPES_SET.difference(typeIDs)
		for typeID in extraTypeIDs:
			v = ret._validators[typeID]
			v._validators.clear()
			v._default = False

	# --------

	return ret
#









